
#define MIN_TESS_LEVEL  3
#define MAX_TESS_LEVEL 32

#define ANGULAR_VELOCITY1 (0.25*PI)
#define ANGULAR_VELOCITY2 (-0.2*PI)

#define MIRRORTXT_W 1024
#define MIRRORTXT_H 1024

#define MIN_ZOOM 0.2
#define MAX_ZOOM 5.0
#define ZOOM_FCT 1.05

typedef struct Camera {
    int    win_width, win_height;
    float  left, right, bottom, top, near, far, rl, tb;
    float  viewer_pos0[4];
    float  viewer_rvec[3];
    double viewer_rangle;
    int    zoomexp;
    double zoom;
  } Camera;

typedef struct Mirror {
    GLuint  mirror_fbo, mirror_txt[2];
    GLuint  mirror_vao, mirror_vbo;
    GLfloat mirror_matrix[16];
  } Mirror;

typedef struct {
    GLuint program_id[2];
    GLint  ColourSourceLoc, colour_source,
           LightingModelLoc, lighting_model,
           NormalSourceLoc, normal_source,
           ModifyDepthLoc, modify_depth;
  } BPRenderPrograms;

typedef struct {
    Camera           camera;
    BezierPatchObjf  *myteapot, *mytorus;
    Mirror           mirror;
    TransBl          trans;
    LightBl          light;
    GLuint           matbuf[2];
    GLuint           mytexture;
    GLint            BezNormals, TessLevel;
    char             cnet, skeleton, animate, vertical, sample_shading, draw_torus;
    float            model_rot_axis[3];
    double           teapot_rot_angle, torus_rot_angle;
    GLfloat          teapot_mmatrix[16], torus_mmatrix[16];
    BPRenderPrograms brprog;
    GLuint           miprog;
  } AppData;


GLuint CreateMyTexture ( int wh );
char LoadMyTextureImage ( GLuint tex, int txwidth, int txheight,
                          int x, int y, const char *filename );
char SetupMyTextureMipmaps ( GLuint tex );

void ConstructMirror ( Mirror *mirror );
void SetupMirrorVPMatrices ( GLfloat eyepos[4], GLfloat reyepos[4],
                             GLfloat mvm[16], GLfloat mpm[16] );
void DrawMirror ( AppData *ad );
void DeleteMirror ( Mirror *mirror );

