#version 450 core

#define MAX_DEG 10

#define MIN_TESS_PARAM 3.0
#define MAX_TESS_PARAM 32.0

layout(local_size_x=1) in;

layout(std430,binding=0) buffer CPoints {
    float cp[];
  } cp;

layout(std430,binding=1) buffer CPIndices {
    int cpi[];
  } cpi;

layout(std430,binding=2) buffer BezPatch {
    int  npatches, dim, udeg, vdeg;
    int  stride_u, stride_v, stride_p, stride_q, nq;
    bool use_ind;
    vec3 Colour;
    int  TessLevel;
    bool BezNormals;
  } bezp;

layout(std430,binding=4) buffer BezPatchTessParams { float tsp[]; };

uniform TransBlock {
    mat4  mm, mmti, vm, pm, vpm;
    vec4  eyepos;
    float left, right, bottom, top, near, far;
    vec4  viewport;
  } trb;

uniform float minedgelgt = 3.0, maxedgelgt = 20.0;
uniform bool clip = false;

int   ncp;
vec4  ctp[(MAX_DEG+1)*(MAX_DEG+1)], bcp[MAX_DEG+1];
float lgtu[3], lgtv[3];

bool ClipOff ( int inst, vec4 pnv )
{
  int i, j;

  for ( i = 0; i < ncp; i++ )
    if ( dot ( ctp[i], pnv ) > 0.0 )
      return false;
  for ( i = 0, j = 12*inst;  i < 6;  i++, j++ )
    tsp[j] = MIN_TESS_PARAM;
  return true;
} /*ClipOff*/

bool NonNegativeZ ( int n, float zw[MAX_DEG+1] )
{
  int i;

  for ( i = 0; i <= n; i++ )
    if ( zw[i] >= 0.0 )
      return true;
  return false;
} /*NonNegativeZ*/

vec4 BCHorner4f ( int n, float t )
{
  int   i, b;
  float s, d;
  vec4  q;

  s = 1.0-t;  d = t;  b = n;
  q = bcp[0];
  for ( i = 1; i <= n; i++ ) {
    q = s*q + (float(b)*d)*bcp[i];
    d *= t;  b = (b*(n-i))/(i+1);
  }
  return q;
} /*BCHorner4f*/

float PolylineLength ( int n, vec2 p[MAX_DEG+1] )
{
  int   i;
  float l;

  l = distance ( p[0], p[1] );
  for ( i = 1; i < n; i++ )
    l += distance ( p[i], p[i+1] );
  return l;
} /*PolylineLength*/

vec2 ProjectP4 ( vec4 p )
{
  vec4 q;

  q = trb.pm * p;
  return vec2 ( trb.viewport.x + 0.5*trb.viewport.z*(q.x/q.w+1.0),
                trb.viewport.y + 0.5*trb.viewport.w*(q.y/q.w+1.0) );
} /*ProjectP4*/

void FindCPolyLengths ( int udeg, int vdeg )
{
  int   i, j, k, m;
  float t, lgt, zw[MAX_DEG+1];
  vec4  p;
  vec2  q[MAX_DEG+1];
#define StoreLGT(DEG,I,TAB) { \
    if ( NonNegativeZ ( DEG, zw ) ) lgt = 1.0e6; \
    else lgt = PolylineLength ( DEG, q ); \
    if ( I == 0 ) TAB[0] = lgt; \
    else if ( I == DEG ) TAB[2] = lgt; \
    if ( lgt > TAB[1] ) TAB[1] = lgt; }

  lgtu[0] = lgtu[1] = lgtu[2] = lgtv[0] = lgtv[1] = lgtv[2] = 0.0;
  for ( i = 0; i <= bezp.udeg; i++ ) {
    t = float(i)/float(bezp.udeg);
    for ( j = 0; j <= bezp.vdeg; j++ ) {
      for ( m = 0, k = j;  m <= bezp.udeg;  m++, k += bezp.vdeg+1 )
        bcp[m] = ctp[k];
      q[j] = ProjectP4 ( p = BCHorner4f ( bezp.udeg, t ) );
      zw[j] = max ( p.z, -p.w );
    }
    StoreLGT ( bezp.udeg, i, lgtv )
  }
  for ( j = 0; j <= bezp.vdeg; j++ ) {
    t = float(j)/float(bezp.vdeg);
    for ( i = k = 0;  i <= bezp.udeg;  i++ ) {
      for ( m = 0;  m <= bezp.vdeg;  m++, k++ )
        bcp[m] = ctp[k];
      q[i] = ProjectP4 ( p = BCHorner4f ( bezp.vdeg, t ) );
      zw[j] = max ( p.z, -p.w );
    }
    StoreLGT ( bezp.vdeg, j, lgtu )
  }
} /*FindCPolyLengths*/

float TessParam ( int deg, float d, float l )
{
  if ( l/d < minedgelgt )
    d = l/minedgelgt;
  l = max ( l/maxedgelgt, d );
  if ( l > MAX_TESS_PARAM ) l = MAX_TESS_PARAM;
  else {
    if ( l < deg ) l = deg;
    if ( l < MIN_TESS_PARAM ) l = MIN_TESS_PARAM;
  }
  return l;
} /*TessParam*/

void FindTessParams ( int inst )
{
  int i, j;

  for ( i = 0, j = 12*inst;  i < 3;  i++, j++ )
    tsp[j] = TessParam ( bezp.udeg, tsp[j+6], lgtu[i] );
  for ( i = 0; i < 3;  i++, j++ )
    tsp[j] = TessParam ( bezp.vdeg, tsp[j+6], lgtv[i] );
} /*FindTessParams*/

void main ( void )
{
  int inst, i, j, k, l, i0;
#define FINDCPOINTS(DIM,Z,W) { \
    if ( bezp.use_ind ) { \
      i0 = inst*bezp.stride_p; \
      for ( i = k = 0;  i <= bezp.udeg;  i++ ) { \
        for ( j = 0;  j <= bezp.vdeg;  j++, k++ ) { \
          l = DIM*cpi.cpi[i0+k]; \
          ctp[k] = trb.vm * (trb.mm * vec4 ( cp.cp[l], cp.cp[l+1], Z, W )); \
    } } } \
    else { \
      i0 = (inst / bezp.nq)*bezp.stride_p+(inst % bezp.nq)*bezp.stride_q; \
      for ( i = k = 0;  i <= bezp.udeg;  i++ ) { \
        for ( j = 0, l = i0+i*bezp.stride_u;  j <= bezp.vdeg; \
              j++, k++, l += bezp.stride_v ) { \
          ctp[k] = trb.vm * (trb.mm * vec4 ( cp.cp[l], cp.cp[l+1], Z, W )); \
    } } } }

  inst = int(gl_GlobalInvocationID.x);
  ncp = (bezp.udeg+1)*(bezp.vdeg+1);
  switch ( bezp.dim ) {
case 2: FINDCPOINTS ( 2, 0.0, 1.0 )  break;
case 3: FINDCPOINTS ( 3, cp.cp[l+2], 1.0 )  break;
case 4: FINDCPOINTS ( 4, cp.cp[l+2], cp.cp[l+3] )  break;
  }
  if ( clip ) {
    if ( ClipOff ( inst, vec4(trb.near, 0.0, trb.left, 0.0) ) )    return;
    if ( ClipOff ( inst, vec4(-trb.near, 0.0, -trb.right, 0.0) ) ) return;
    if ( ClipOff ( inst, vec4(0.0, trb.near, trb.bottom, 0.0) ) )  return;
    if ( ClipOff ( inst, vec4(0.0, -trb.near, -trb.top, 0.0) ) )   return;
    if ( ClipOff ( inst, vec4(0.0, 0.0, -1.0, -trb.near) ) )       return;
    if ( ClipOff ( inst, vec4(0.0, 0.0, 1.0, trb.far) ) )          return;
  }
  FindCPolyLengths ( bezp.udeg, bezp.vdeg );
  FindTessParams ( inst );
} /*main*/
