
#define MIN_TESS_LEVEL  3
#define MAX_TESS_LEVEL 32

#define ANGULAR_VELOCITY1 (0.25*PI)

#define MIRRORTXT_W 1024
#define MIRRORTXT_H 1024

typedef struct Camera {
    int    win_width, win_height;
    float  left, right, bottom, top, near, far, rl, tb;
    float  viewer_pos0[4];
    float  viewer_rvec[3];
    double viewer_rangle;
  } Camera;

typedef struct Mirror {
    GLuint  mirror_fbo, mirror_txt[2];
    GLuint  mirror_vao, mirror_vbo;
    GLfloat mirror_matrix[16];
  } Mirror;

typedef struct {
    GLuint program_id[5];
    GLint  ColourSourceLoc, colour_source,
           LightingModelLoc, lighting_model,
           NormalSourceLoc, normal_source,
           ModifyDepthLoc, modify_depth;
  } BPRenderPrograms;

typedef struct {
    GLuint program_id;
    GLuint dim_loc, trnum_loc;
    GLuint cpibp, cpobp, ctrbp, ctribp;
  } KLArticulationProgram;

typedef struct {
    BezierPatchObjf *bpatches[2];
    GLuint          texture, matbuf, tribuf;
    GLint           ntr, mtn;
  } KLBezPatches;

typedef struct {
    Camera                camera;
    kl_linkage            *linkage;
    KLBezPatches          bezp[2];
    Mirror                mirror;
    TransBl               trans;
    LightBl               light;
    MatBl                 mat;
    GLuint                lktrbuf;
    GLint                 BezNormals, TessLevel;
    char                  cnet, skeleton, shadows, animate, final;
    double                teapot_rot_angle;
    BPRenderPrograms      brprog;
    GLuint                miprog[2];
    KLArticulationProgram artprog;
  } AppData;

void SetModelMatrix ( AppData *ad, GLfloat mm[16] );

void LoadLinkageArticulationProgram ( KLArticulationProgram *prog );
void DeleteLinkageArticulationProgram ( KLArticulationProgram *prog );
kl_linkage *ConstructMyLinkage ( AppData *ad );
void ArticulateMyLinkage ( kl_linkage *lk );
void DrawMyLinkage ( AppData *ad, char final );

void LoadMirrorShaders ( GLuint program_id[2] );
void ConstructMirror ( Mirror *mirror );
void SetupMirrorVPMatrices ( TransBl *trans,
                             GLfloat eyepos[4], GLfloat reyepos[4],
                             GLfloat mvm[16], GLfloat mpm[16] );
void DrawMirror ( Mirror *mirror, GLuint program_id, char final );
void DeleteMirror ( Mirror *mirror );

