#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include "../utilities/openglheader.h"

#include "../utilities/utilities.h"
#include "trans.h"

#define NTRUOFFS 13

static GLuint trbbp = GL_INVALID_INDEX;
static GLint  trbsize, trbofs[NTRUOFFS];
static const GLchar *UTBNames[] =
  { "TransBlock", "TransBlock.mm", "TransBlock.mmti", "TransBlock.vm",
    "TransBlock.pm", "TransBlock.vpm", "TransBlock.eyepos",
    "TransBlock.left", "TransBlock.right", "TransBlock.bottom",
    "TransBlock.top", "TransBlock.near", "TransBlock.far",
    "TransBlock.viewport" };

GLuint GetAccessToTransBlockUniform ( GLuint program_id )
{
  if ( trbbp == GL_INVALID_INDEX )
    GetAccessToUniformBlock ( program_id, NTRUOFFS, &UTBNames[0],
                              &trbsize, trbofs, &trbbp );
  else
    AttachUniformBlockToBP ( program_id, UTBNames[0], trbbp );
  return trbbp;
} /*GetAccessToTransBlockUniform*/

GLuint NewUniformTransBlock ( void )
{
  return NewUniformBuffer ( trbsize, trbbp );
} /*NewUniformTransBlock*/

void AttachUniformTransBlockToBP ( GLuint program_id )
{
  AttachUniformBlockToBP ( program_id, UTBNames[0], trbbp );
} /*AttachUniformTransBlockToBP*/

void ConstructPMatrix ( TransBl *trans, float left, float right,
                        float bottom, float top, float near, float far,
                        GLint viewport[4], char mirror )
{
  GLfloat *pm, *fp;

  if ( mirror )
    pm = trans->mpm;
  else {
    pm = trans->wpm;
    fp = trans->wfp;
    fp[0] = left;  fp[1] = right;  fp[2] = bottom;
    fp[3] = top;   fp[4] = near;   fp[5] = far;
    memcpy ( trans->viewport, viewport, 4*sizeof(GLint) );
  }
  M4x4Frustumf ( pm, NULL, left, right, bottom, top, near, far );
  ExitIfGLError ( "ConstructPMatrix" );
} /*ConstructPMatrix*/

void LoadVPMatrix ( TransBl *trans, char mirror )
{
  GLfloat *vm, *pm, *ep, *fp, vp[4], vpm[16];
  int     i;

  glBindBuffer ( GL_UNIFORM_BUFFER, trans->trbuf );
  if ( mirror ) {
    vm = trans->mvm;  pm = trans->mpm;  ep = trans->reyepos;
  }
  else {
    fp = trans->wfp;
    glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[6], sizeof(GLfloat), &fp[0] );
    glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[7], sizeof(GLfloat), &fp[1] );
    glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[8], sizeof(GLfloat), &fp[2] );
    glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[9], sizeof(GLfloat), &fp[3] );
    glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[10], sizeof(GLfloat), &fp[4] );
    glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[11], sizeof(GLfloat), &fp[5] );
    for ( i = 0; i < 4; i++ )
      vp[i] = (GLfloat)trans->viewport[i];
    glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[12], 4*sizeof(GLfloat), vp );
    vm = trans->wvm;  pm = trans->wpm;  ep = trans->eyepos;
  }
  M4x4Multf ( vpm, pm, vm );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[2], 16*sizeof(GLfloat), vm );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[3], 16*sizeof(GLfloat), pm );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[4], 16*sizeof(GLfloat), vpm );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[5], 4*sizeof(GLfloat), ep );
  ExitIfGLError ( "LoadVPMatrix" );
} /*LoadPMatrix*/

void LoadMMatrix ( TransBl *trans, GLfloat mm[16] )
{
  GLfloat mmti[16];

  if ( mm )
    memcpy ( trans->mm, mm, 16*sizeof(GLfloat) );
  glBindBuffer ( GL_UNIFORM_BUFFER, trans->trbuf );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[0], 16*sizeof(GLfloat), trans->mm );
  M4x4TInvertf ( mmti, trans->mm );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[1], 16*sizeof(GLfloat), mmti );
  ExitIfGLError ( "LoadMMatrix" );
} /*LoadMMatrix*/

void LoadShTrans ( TransBl *trans, GLfloat vm[16], GLfloat pm[16], GLfloat eyepos[4] )
{
  GLfloat vpm[16];

  M4x4Multf ( vpm, pm, vm );
  glBindBuffer ( GL_UNIFORM_BUFFER, trans->trbuf );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[2], 16*sizeof(GLfloat), vm );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[3], 16*sizeof(GLfloat), pm );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[4], 16*sizeof(GLfloat), vpm );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[5], 4*sizeof(GLfloat), eyepos );
  ExitIfGLError ( "LoadShTrans" );
} /*LoadShTrans*/

