#include <stdlib.h>
#include <stdio.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "openglheader.h"
#include <GL/glx.h>

#include "utilities.h"
#include "initglxctx.h"
#include "app1.h"

Window     xmywin;
GLXContext glxcontext;
Colormap   xcolormap;
char       terminate;

int opcja = 2;

void DestroyMyShaders ( void )
{
} /*DestroyMyShaders*/

void DestroyMyObject ( void )
{
} /*DestroyMyObject*/

void InitMyGLXWindow ( int argc, char **argv,
                       int major, int minor, int width, int height )
{
  XSetWindowAttributes swa;
  XVisualInfo          *xvii;

  InitGLXContext ( major, minor, 0, NULL, &xvii, &glxcontext );
  if ( !(xcolormap = XCreateColormap ( xdisplay, xrootwin,
                                       xvii->visual, AllocNone )) )
    exit ( 1 );
  swa.colormap = xcolormap;
  swa.event_mask = ExposureMask | StructureNotifyMask| ButtonPressMask |
                   ButtonReleaseMask | PointerMotionMask | KeyPressMask ;
  xmywin = XCreateWindow ( xdisplay, xrootwin, 0, 0, width, height,
                           0, xvii->depth, InputOutput, xvii->visual,
                           CWColormap | CWEventMask, &swa );
  XFree ( xvii );
  XSetWMProtocols ( xdisplay, xmywin, &DeleteWindow, 1 );
  XMapWindow ( xdisplay, xmywin );
  if ( !glXMakeCurrent ( xdisplay, xmywin, glxcontext ) ) {
    printf ( "Error: glXMakeCurrent failed\n" );
    exit ( 1 );
  }
  GetGLProcAddresses ( major, minor );
  PrintGLVersion ();
} /*InitMyGLXWindow*/

void DestroyMyGLXWindow ( void )
{
  XDestroyWindow ( xdisplay, xmywin );
  XFreeColormap ( xdisplay, xcolormap );
} /*DestroyMyGLXWindow*/

void Initialise ( int argc, char **argv )
{
  InitXServerConnection ( argc, argv, false );
  InitMyGLXWindow ( argc, argv, 4, 2, 480, 360 );
  InitMyObject ();
} /*Initialise*/

void MyWinExpose ( void )
{
  glClearColor ( 1.0, 1.0, 1.0, 1.0 );
  glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
  glEnable ( GL_DEPTH_TEST );
  DrawIcosahedron ( opcja );
  glFlush ();
  glXSwapBuffers ( xdisplay, xmywin );
} /*MyWinExpose*/

void MyWinConfigureNotify ( XEvent *event )
{
  int     width, height;
  GLfloat m[16];
  float   lr;

  width = event->xconfigure.width;
  height = event->xconfigure.height;
  glViewport ( 0, 0, width, height );  /* klatka jest calym oknem */
  lr = 0.5533*(float)width/(float)height;
  M4x4Frustumf ( m, NULL, -lr, lr, -0.5533, 0.5533, 5.0, 20.0 );
  glBindBuffer ( GL_UNIFORM_BUFFER, trbuf );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[2], 16*sizeof(GLfloat), m );
  ExitIfGLError ( "ReshapeFunc" );
} /*MyWinConfigureNotify*/

void MyWinButtonPress ( XEvent *event )
{
} /*MyWinButtonPress*/

void MyWinButtonRelease ( XEvent *event )
{
} /*MyWinButtonRelease*/

void MyWinMotionNotify ( XEvent *event )
{
} /*MyWinMotionNotify*/

void MyWinKeyPress ( XEvent *event )
{
  terminate = 1;
} /*MyWinKeyPress*/

void MyWinClientMessage ( XClientMessageEvent *xclient )
{
  if ( xclient->message_type == WMProtocols &&
       (Atom)xclient->data.l[0] == DeleteWindow ) {
    terminate = true;
    return;
  }
} /*MyWinClientMessage*/

void MyWinMessageProc ( XEvent *event )
{
  switch ( event->xany.type ) {
case Expose:
    if ( event->xexpose.count == 0 )
      MyWinExpose ();
    break;
case ConfigureNotify:  MyWinConfigureNotify ( event );  break;
case ButtonPress:      MyWinButtonPress ( event );      break;
case ButtonRelease:    MyWinButtonRelease ( event );    break;
case MotionNotify:     MyWinMotionNotify ( event );     break;
case KeyPress:         MyWinKeyPress ( event );         break;
case ClientMessage: MyWinClientMessage (
                        (XClientMessageEvent*)event );  break;
           default:    break;
  }
} /*MyWinMessageProc*/

void MessageLoop ( void )
{
  XEvent event;

  terminate = 0;
  do {
    XNextEvent ( xdisplay, &event );
    if ( event.xany.window == xmywin )
      MyWinMessageProc ( &event );
  } while ( !terminate );
} /*MessageLoop*/

void Cleanup ( void )
{
  glUseProgram ( 0 );
  glDeleteProgram ( program_id );
  glDeleteBuffers ( 1, &trbuf );
  glDeleteVertexArrays ( 1, &icos_vao );
  glDeleteBuffers ( 3, icos_vbo );
  ExitIfGLError ( "Cleanup" );
  DestroyMyGLXWindow ();
  XCloseDisplay ( xdisplay );
} /*Cleanup*/

int main ( int argc, char **argv )
{
  Initialise ( argc, argv );
  MessageLoop ();
  Cleanup ();
  exit ( 0 );
} /*main*/
