#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "openglheader.h"
#include <GL/glx.h>

#include "utilities.h"
#include "initglxctx.h"
#include "xwidgets.h"
#include "lights.h"
#include "trans.h"
#include "app1b-glx.h"

char        terminate;
Window      window[3];
GLXContext  glxcontext;
Colormap    xcolormap;
XVisualInfo *xvii;

int          window0_width, window0_height;
xwinmenu    *wm1, *wm2;
my3Dwidget  *mywdg;
myglwindata  myglwdata;

int opti = 0;

char str_EXIT[]     = "Exit";
char str_FACETS[]   = "facets";
char str_EDGES[]    = "edges";
char str_VERTICES[] = "vertices";
char str_LIGHT[]    = "light";
char str_ANIMATE[]  = "animate";

char  swfac = true, swedg = false, swvert = false, swlight = true;
float speed = 0.0;


void Win1Callback ( struct xwidget *wdg, int msg, int key, int x, int y )
{
  switch ( msg ) {
case WDGMSG_BUTTON_PRESS:
    switch ( wdg->id ) {
  case BTN_ID_EXIT:
      terminate = true;
      break;
  default:
      break;
    }
    break;

case WDGMSG_SWITCH_CHANGE:
    switch ( wdg->id ) {
  case SW_ID_FACETS:
      if ( swfac ) { swedg = swvert = false; }
              else { swedg = true;  swvert = false; }
      goto let_redraw;
  case SW_ID_EDGES:
      if ( swedg ) { swvert = swfac = false; }
              else { swvert = true;  swfac = false; }
      goto let_redraw;
  case SW_ID_VERTICES:
      if ( swvert ) { swfac = swedg = false; }
               else { swfac = true;  swedg = false; }
let_redraw:
      if ( swvert ) myglwdata.model_option = 0;
      else myglwdata.model_option = swedg ? 1 : 2;
      wm2->changed = true;
      PostExposeEvent ( window[0], window0_width, window0_height );
      break;
  case SW_ID_LIGHT:
      myglwdata.light_option = swlight;
      wm2->changed = true;
      PostExposeEvent ( window[2], wm2->r.width, wm2->r.height );
      break;
  case SW_ID_ANIMATE:
      ToggleAnimation ( &myglwdata, myglwdata.animation );
      break;
  default:
      break;
    }
    break;

case WDGMSG_SLIDEBAR_CHANGE:
    switch ( wdg->id ) {
  case SL_ID_SPEED:
      myglwdata.speed = 3.1415926*(0.1+0.9*speed);
      if ( myglwdata.animation )
        TimerTic ();
      break;
  default:
      break;
    }
    break;

default:
    break;
  }
} /*Win1Callback*/

xwinmenu *SetupAppMenu ( void )
{
  xwinmenu *wm;

  if ( !(wm = NewWinMenu ( window[1], MENU_WIDTH, WIN0_HEIGHT, 0, 0,
                           NULL, NULL, Win1Callback )) )
    ExitOnError ( "SetupAppMenu" );
  NewButton ( wm, BTN_ID_EXIT, 60, 18, 2, 2, str_EXIT );
  NewSwitch ( wm, SW_ID_FACETS, 60, 16, 2, 22, str_FACETS, &swfac );
  NewSwitch ( wm, SW_ID_EDGES, 60, 16, 2, 42, str_EDGES, &swedg );
  NewSwitch ( wm, SW_ID_VERTICES, 60, 16, 2, 62, str_VERTICES, &swvert );
  NewSwitch ( wm, SW_ID_LIGHT, 60, 16, 2, 86, str_LIGHT, &swlight );
  NewSwitch ( wm, SW_ID_ANIMATE, 60, 16, 2, 110, str_ANIMATE, &myglwdata.animation );
  NewSlidebarf ( wm, SL_ID_SPEED, 110, 10, 2, 130, &speed );
  return wm;
} /*SetupAppMenu*/

void AdvanceAnimation ( void )
{
  if ( myglwdata.animation ) {
    if ( (model_rot_angle += myglwdata.speed * TimerTocTic ()) >= PI )
      model_rot_angle -= 2.0*PI;
    SetupModelMatrix ( model_rot_axis, model_rot_angle );
    wm2->changed = true;
    PostClientMessageEvent ( window[2], CLIENTMSG_ANIMATE, 8, NULL );
    PostExposeEvent ( window[2], wm2->r.width, wm2->r.height );
  }
} /*AdvanceAnimation*/

void ToggleAnimation ( myglwindata *wd, char on )
{
  if ( (wd->animation = on) ) {
    TimerTic ();
    PostClientMessageEvent ( window[2], CLIENTMSG_ANIMATE, 8, NULL );
  }
  wm1->changed = true;
  PostExposeEvent ( window[0], window0_width, window0_height );
} /*ToggleAnimation*/

char My3DWidgetInput ( struct xwidget *wdg, int msg, int key, int x, int y )
{
  myglwindata *wd;

  wd = (myglwindata*)wdg->data0;
  switch ( wdg->state ) {
case WDGSTATE_DEFAULT:
    switch ( msg ) {
  case XWMSG_BUTTON_PRESS:
      if ( key == Button1 ) {
        wd->prevx = x;  wd->prevy = y;
        wdg->state = WDGSTATE_VIEW_TURNING;
        GrabInput ( wdg );
        return true;
      }
      break;
  case XWMSG_KEY_PRESS:
      goto process_key;
  default:
      break;
    }
    break;

case WDGSTATE_VIEW_TURNING:
    switch ( msg ) {
  case XWMSG_MOUSE_MOTION:
      if ( ((XMotionEvent*)wdg->wm->ev)->state & Button1Mask ) {
        RotateViewer ( x - wd->prevx, y - wd->prevy );
        wd->prevx = x;  wd->prevy = y;
        wdg->wm->changed = true;
      }
      else
        goto release;
      break;
  case XWMSG_BUTTON_RELEASE:
      if ( key == Button1 ) {
release:
        wdg->state = WDGSTATE_DEFAULT;
        UngrabInput ( wdg );
        return true;
      }
      break;
  case XWMSG_KEY_PRESS:
process_key:
      switch ( key ) {
    case ' ':
        ToggleAnimation ( wd, !wd->animation );
        return true;
    default:
        break;
      }
      break;
  default:
      break;
    }
    break;

default:
    break;
  }
  return false;
} /*My3DWidgetInput*/

void My3DWidgetRedraw ( struct xwidget *wdg )
{
  XRectangle *r;
  my3Dwidget *w;

  r = &wdg->wm->r;
  w = (my3Dwidget*)wdg;
  glViewport ( 0, 0, r->width, r->height );
  w->redraw ( w );
} /*My3DWidgetRedraw*/

my3Dwidget *New3DWidget ( struct xwinmenu *wm, int id,
                          int w, int h, int x, int y,
                          my3Dredraw redraw, void *data0, void *data1 )
{
  my3Dwidget *wdg;

  if ( (wdg = (my3Dwidget*)NewWidget ( wm, sizeof(my3Dwidget), id, w, h, x, y,
                              My3DWidgetInput, My3DWidgetRedraw, data0, data1 )) ) {
    wdg->redraw = redraw;
  }
  return wdg;
} /*New3DWidget*/

void RedrawWin2 ( xwinmenu *wm )
{
  xwidget *wdg;

  for ( wdg = wm->wlist.next;  wdg;  wdg = wdg->link.next )
    wdg->redraw ( wdg );
  glFlush ();
  glXSwapBuffers ( xdisplay, wm->window );
  if ( opti > 0 ) {
    opti --;
    PostExposeEvent ( wm->window, wm->r.width, wm->r.height );
  }
} /*RedrawWin2*/

void Win2Reshape ( xwinmenu *wm, short w, short h )
{
  myglwindata *wd;

  wd = (myglwindata*)mywdg->wdg.data0;
  mywdg->wdg.r.width = w;  mywdg->wdg.r.height = h;
  wd->bottom = -(wd->top = 0.5533);
  wd->left = -(wd->right = 0.5533*(float)w/(float)h );
  wd->near = 5.0;  wd->far = 15.0;
  wd->rl = wd->right - wd->left;  wd->tb = wd->top - wd->bottom;
  M4x4Frustumf ( trans.pm, NULL, wd->left, wd->right, wd->bottom, wd->top,
                 wd->near, wd->far );
  LoadVPMatrix ( &trans );
  opti = 3;
  PostExposeEvent ( wm->window, w, h );
  ExitIfGLError ( "Win2Reshape" );
} /*Win2Reshape*/

void Win2Callback ( struct xwidget *wdg, int msg, int key, int x, int y )
{
  switch ( msg ) {
case WDGMSG_RECONFIGURE:
    Win2Reshape ( wdg->wm, x, y );
    break;
case XWMSG_CLIENT_MESSAGE:
    if ( key == CLIENTMSG_ANIMATE )
      AdvanceAnimation ();
    break;
default:
    break;
  }
} /*Win2Callback*/

void App3Redraw ( my3Dwidget *wdg )
{
  glClearColor ( 1.0, 1.0, 1.0, 1.0 );
  glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
  glEnable ( GL_DEPTH_TEST );
  DrawIcosahedron ( myglwdata.model_option, myglwdata.light_option );
} /*App3Redraw*/

xwinmenu *SetupAppGLWindow ( void )
{
  xwinmenu *wm;

  if ( !(wm = NewWinMenu ( window[2], WIN0_WIDTH-MENU_WIDTH, WIN0_HEIGHT, 0, 0,
                           NULL, RedrawWin2, Win2Callback )) )
    ExitOnError ( "SetupAppGLWindow 0" );
  if ( !(mywdg = New3DWidget ( wm, GLWIN_ID_VIEW, wm->r.width, wm->r.height, 0, 0,
                               App3Redraw, &myglwdata, NULL )) )
    ExitOnError ( "SetupAppGLWindow 1" );
  return wm;
} /*SetupAppGLWindow*/

void InitMyWindows ( int argc, char **argv )
{
  XSetWindowAttributes swa;
  static const int     wx[3] = { 0, 0, MENU_WIDTH };
  static const int     wh[3] = { WIN0_WIDTH, MENU_WIDTH, WIN0_WIDTH-MENU_WIDTH };
  Window               upw;
  int                  i;

  InitGLXContext ( 4, 2, 0, NULL, &xvii, &glxcontext );
  if ( !(xcolormap = XCreateColormap ( xdisplay, xrootwin,
                                       xvii->visual, AllocNone )) )
    ExitOnError ( "InitMyGLXWindow 0" );
  swa.colormap = xcolormap;
  swa.event_mask = ExposureMask | StructureNotifyMask| ButtonPressMask |
                   ButtonReleaseMask | PointerMotionMask | KeyPressMask;
  for ( i = 0, upw = xrootwin; i < 3; i++, upw = window[0] ) {
    window[i] = XCreateWindow ( xdisplay, upw, wx[i], 0, wh[i], WIN0_HEIGHT,
                             0, xvii->depth, InputOutput, xvii->visual,
                             CWColormap | CWEventMask, &swa );
    XMapWindow ( xdisplay, window[i] );
  }
  XSetWMProtocols ( xdisplay, window[0], &DeleteWindow, 1 );
  XStoreName ( xdisplay, window[0], "app1b-glx" );
  xgc = XCreateGC ( xdisplay, window[1], 0, 0 );
  InitRGBXColourmap ( xvii );
  InitWinMenuPalette ();
  if ( !glXMakeCurrent ( xdisplay, window[2], glxcontext ) )
    ExitOnError ( "InitMyWindows 2" );
  GetGLProcAddresses ( 4, 2 );
  PrintGLVersion ();
  window0_width = WIN0_WIDTH;
  window0_height = WIN0_HEIGHT;
} /*InitMyWindows*/

void Initialise ( int argc, char **argv )
{
  InitXServerConnection ( argc, argv, false );
  InitMyWindows ( argc, argv );
  wm1 = SetupAppMenu ();
  wm2 = SetupAppGLWindow ();
  InitMyObject ();
  Win2Reshape ( wm2, wm2->r.width, wm2->r.height );
} /*Initialise*/

/* ////////////////////////////////////////////////////////////////////////// */
void Win0ConfigureNotify ( int width, int height )
{
  window0_width  = width;
  window0_height = height;
  XMoveResizeWindow ( xdisplay, window[1], 0, 0, MENU_WIDTH, height );
  XMoveResizeWindow ( xdisplay, window[2], MENU_WIDTH, 0, width-MENU_WIDTH, height );
} /*Win0ConfigureNotify*/

void Win0ClientMessage ( XClientMessageEvent *xclient )
{
  if ( xclient->message_type == WMProtocols &&
       (Atom)xclient->data.l[0] == DeleteWindow ) {
    terminate = true;
    return;
  }
} /*Win0ClientMessage*/

void Win0MessageProc ( XEvent *ev )
{
  switch ( ev->xany.type ) {
case Expose:
    PostExposeEvent ( wm1->window, wm1->r.width, wm1->r.height );
    PostExposeEvent ( wm2->window, wm2->r.width, wm2->r.height );
    break;
case ConfigureNotify:
    Win0ConfigureNotify ( ev->xconfigure.width, ev->xconfigure.height );
    break;
case ClientMessage:
    Win0ClientMessage ( &ev->xclient );
    break;
default:
    break;
  }
} /*Win0MessageProc*/

/* ////////////////////////////////////////////////////////////////////////// */
void MessageLoop ( void )
{
  XEvent ev;

  terminate = 0;
  do {
    XNextEvent ( xdisplay, &ev );
    if ( ev.xany.window == window[0] )
      Win0MessageProc ( &ev );
    else if ( ev.xany.window == window[1] )
      WinMenuInput ( wm1, &ev );
    else if ( ev.xany.window == window[2] )
      WinMenuInput ( wm2, &ev );
  } while ( !terminate );
} /*MessageLoop*/

void Cleanup ( void )
{
  DestroyMyObject ();
  DestroyMyShaders ();
  DeleteWinMenu ( wm1 );
  DeleteWinMenu ( wm2 );
  XDestroySubwindows ( xdisplay, window[0] );
  XDestroyWindow ( xdisplay, window[0] );
  XFreeGC ( xdisplay, xgc );
  XFree ( xvii );
  XFreeColormap ( xdisplay, xcolormap );
  XCloseDisplay ( xdisplay );
} /*Cleanup*/

int main ( int argc, char **argv )
{
/*setvbuf ( stdout, NULL, _IONBF, 0 );*/
  Initialise ( argc, argv );
  MessageLoop ();
  Cleanup ();
  exit ( 0 );
} /*main*/

