#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/times.h>
#include "openglheader.h"
#include <GLFW/glfw3.h>

#include "utilities.h"
#include "trans.h"
#include "lights.h"
#include "bspatches.h"
#include "app2c.h"

GLFWwindow   *mywindow;
int     win_width, win_height;
double  last_xi, last_eta;
float   left, right, bottom, top, near, far;

int     app_state = STATE_NOTHING;
float   viewer_rvec[3] = {1.0,0.0,0.0};
float   viewer_rangle = 0.0;
const   float viewer_pos0[4] = {0.0,0.0,10.0,1.0};

GLint   BSNormals = GL_TRUE;
GLint   TessLevel = 30;
char    redraw;
char    animate = 0;
clock_t zero_time, start_time;
float   clocks_per_sec;
int     opti = 0;

void myGLFWErrorHandler ( int error, const char *description )
{
  fprintf ( stderr, "GLFW error: %s\n", description );
  exit ( 1 );
} /*myGLFWErrorHandler*/

void ReshapeFunc ( GLFWwindow *win, int width, int height )
{
  float lr;

  glViewport ( 0, 0, width, height );
  lr = 0.5*(float)width/(float)height;
  M4x4Frustumf ( trans.pm, NULL, -lr, lr, -0.5, 0.5, 5.0, 20.0 );
  LoadVPMatrix ( &trans );
  win_width = width,  win_height = height;
  left = -(right = lr);  bottom = -(top = 0.5);  near = 5.0;  far = 20.0;
  ExitIfGLError ( "ReshapeFunc" );
  redraw = 1;
  opti = 4;
} /*ReshapeFunc*/

void Redraw ( GLFWwindow *win )
{
  glClearColor ( 1.0, 1.0, 1.0, 1.0 );
  glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
  glEnable ( GL_DEPTH_TEST );
  SetModelMatrix ( patch_mmatrix, patch_mmti );
  DrawMyPatch ();
  glUseProgram ( 0 );
  glFlush ();
  glfwSwapBuffers ( win );
  ExitIfGLError ( "Redraw" );
  redraw = 0;
} /*Redraw*/

void DisplayFunc ( GLFWwindow *win )
{
  redraw = 1;
} /*DisplayFunc*/

void ToggleAnimation ( void )
{
  struct tms clk;

  if ( (animate = !animate) )
    start_time = times ( &clk );
  else
    patch_rot_angle0 = patch_rot_angle;
} /*ToggleAnimation*/

void KeyFunc ( GLFWwindow *win, int key, int scancode, int action, int mode )
{
  if ( action == GLFW_PRESS || action == GLFW_REPEAT ) {
    switch ( key ) {
  case GLFW_KEY_ESCAPE:
      glfwSetWindowShouldClose ( mywindow, 1 );
      break;
  default:
      break;
    }
  }
} /*KeyFunc*/

void CharFunc ( GLFWwindow *win, unsigned int charcode )
{
  switch ( charcode ) {
case '+':
    if ( TessLevel < MAX_TESS_LEVEL ) {
      SetBSPatchTessLevel ( mypatch, ++TessLevel );
      redraw = 1;
    }
    break;
case '-':
    if ( TessLevel > MIN_TESS_LEVEL ) {
      SetBSPatchTessLevel ( mypatch, --TessLevel );
      redraw = 1;
    }
    break;
case 'N':  case 'n':
    BSNormals = BSNormals == 0;
    SetBSPatchNormals ( mypatch, BSNormals );
    redraw = 1;
    break;
case ' ':
    ToggleAnimation ();
    break;
case 'C':  case 'c':
    cnet = !cnet;
    redraw = 1;
    break;
case 'L':  case 'l':
    skeleton = 1;
    redraw = 1;
    break;
case 'S':  case 's':
    skeleton = 0;
    redraw = 1;
    break;
case 'B':  case 'b':
    glUseProgram ( progid[0] );
    glUniform1i ( LightingModelLoc, lighting_model = !lighting_model );
    redraw = 1;
    break;
case '0':
    glUseProgram ( progid[0] );
    glUniform1i ( ColourSourceLoc, 0 );
    redraw = 1;
    break;
case '1':
    glUseProgram ( progid[0] );
    glUniform1i ( ColourSourceLoc, 1 );
    redraw = 1;
    break;
default:    /* ignorujemy wszystkie inne klawisze */
    break;
  }
} /*CharFunc*/

void MouseFunc ( GLFWwindow *win, int button, int action, int mods )
{
  switch ( app_state ) {
case STATE_NOTHING:
    if ( button == GLFW_MOUSE_BUTTON_LEFT && action == GLFW_PRESS ) {
      glfwGetCursorPos ( win, &last_xi, &last_eta );
      app_state = STATE_TURNING;
    }
    break;
case STATE_TURNING:
    if ( button == GLFW_MOUSE_BUTTON_LEFT && action != GLFW_PRESS )
      app_state = STATE_NOTHING;
    break;
default:
    break;
  }
} /*MouseFunc*/

void MotionFunc ( GLFWwindow *win, double x, double y )
{
  switch ( app_state ) {
case STATE_TURNING:
    if ( x != last_xi || y != last_eta ) {
      RotateViewer ( x-last_xi, y-last_eta );
      last_xi = x,  last_eta = y;
      redraw = 1;
    }
    break;
default:
    break;
  }
} /*MotionFunc*/

void IdleFunc ( void )
{
  struct tms clk;
  clock_t    tt;
  float      deltat;

  tt = times ( &clk );
  deltat = (float)(tt-start_time);
  patch_rot_angle = patch_rot_angle0 + 0.78539816*deltat/clocks_per_sec;
  SetupPatchMatrix ();
  redraw = 1;
} /*IdleFunc*/

void Initialise ( int argc, char **argv )
{
  glfwSetErrorCallback ( myGLFWErrorHandler );
  if ( !glfwInit () ) {
    fprintf ( stderr, "Error: glfwInit failed\n" );
    exit ( 1 );
  }
  if ( !(mywindow = glfwCreateWindow ( 480, 360,
                                       "powierzchnia B-sklejana", NULL, NULL )) ) {
    glfwTerminate ();
    fprintf ( stderr, "Error: glfwCreateWindow failed\n" );
    exit ( 1 );
  }
  glfwMakeContextCurrent ( mywindow );
  GetGLProcAddresses ( APP2C_GL_MAJOR, APP2C_GL_MINOR );
  PrintGLVersion ();
  glfwSetWindowSizeCallback ( mywindow, ReshapeFunc );
  glfwSetWindowRefreshCallback ( mywindow, DisplayFunc );
  glfwSetKeyCallback ( mywindow, KeyFunc );
  glfwSetCharCallback ( mywindow, CharFunc );
  glfwSetMouseButtonCallback ( mywindow, MouseFunc );
  glfwSetCursorPosCallback ( mywindow, MotionFunc );
  InitMyObject ();
  ReshapeFunc ( mywindow, 480, 360 );
  redraw = 1;
} /*Initialise*/

void Cleanup ( void )
{
  int i;

  glUseProgram ( 0 );
  for ( i = 0; i < 1; i++ )
    glDeleteProgram ( progid[i] );
  glDeleteBuffers ( 1, &trans.trbuf );
  glDeleteBuffers ( 1, &light.lsbuf );
  DeleteBSplinePatch ( mypatch );
  DeleteEmptyVAO ();
  glDeleteBuffers ( 2, matbuf );
  glfwDestroyWindow ( mywindow );
  glfwTerminate ();
} /*Cleanup*/

static void (*idlefunc)(void) = NULL;

void SetIdleFunc ( void(*IdleFunc)(void) )
{
  idlefunc = IdleFunc;
} /*SetIdleFunc*/

void MessageLoop ( void )
{
  do {
    if ( idlefunc ) {
      idlefunc ();
      glfwPollEvents ();
    }
    else
      glfwWaitEvents ();
    if ( redraw )
      Redraw ( mywindow );
  } while ( !glfwWindowShouldClose ( mywindow ) );
} /*MessageLoop*/

int main ( int argc, char **argv )
{
  Initialise ( argc, argv );
  SetIdleFunc ( IdleFunc );
  MessageLoop ();
  Cleanup ();
  exit ( 0 );
} /*main*/

