#version 420 core

#define MAX_NLIGHTS 8

layout(early_fragment_tests) in;

in FVertex {
    vec3 Colour;
    vec3 Position;
    vec3 Normal, TNormal;
  } In;

out vec4 out_Colour;

uniform TransBlock {
    mat4 mm, mmti, vm, pm, vpm;
    vec4 eyepos;
  } trb;

struct LSPar {
    vec4 position;
    vec3 ambient;
    vec3 direct;
    vec3 attenuation;
  };

uniform LSBlock {
    uint  nls;              /* liczba zrodel swiatla */
    uint  mask;             /* maska wlaczonych zrodel */
    LSPar ls[MAX_NLIGHTS];  /* poszczegolne zrodla swiatla */
  } light;

struct Material {
    vec3  ambref;     /* odbijanie swiatla rozproszonego */
    vec3  dirref;     /* odbijanie swiatla kierunkowego */
    vec3  specref;    /* odbijanie zwierciadlane swiatla */
    float shininess, wa, we;  /* polysk */
  };

uniform MatBlock {
    Material mat;
  } mat;

uniform int ColourSource = 0, LightingModel = 0;

Material mm;
vec3     normal, tnormal;

void GetColours ( void )
{
  switch ( ColourSource ) {
default:
    mm.ambref = mm.dirref = In.Colour;
    mm.specref = vec3 ( 0.25 );
    mm.shininess = 20.0;  mm.wa = mm.we = 1.0;
    break;
case 1:
    mm = mat.mat;
    break;
  }
} /*GetColours*/

vec3 posDifference ( vec4 p, vec3 pos, out float dist )
{
  vec3 v;

  if ( p.w != 0.0 ) {
    v = p.xyz/p.w-pos.xyz;
    dist = sqrt ( dot ( v, v ) );
  }
  else
    v = p.xyz;
  return normalize ( v );
} /*posDifference*/

float attFactor ( vec3 att, float dist )
{
  return 1.0/(((att.z*dist)+att.y)*dist+att.x);
} /*attFactor*/

vec3 LambertLighting ( void )
{
  vec3  lv, vv, Colour;
  float d, e, dist;
  uint  i, mask;

  vv = posDifference ( trb.eyepos, In.Position, dist );
  e = dot ( vv, tnormal );
  Colour = vec3(0.0);
  for ( i = 0, mask = 0x00000001;  i < light.nls;  i++, mask <<= 1 )
    if ( (light.mask & mask) != 0 ) {
      Colour += light.ls[i].ambient * mm.ambref;
      lv = posDifference ( light.ls[i].position, In.Position, dist );
      d = dot ( lv, normal );
      if ( e > 0.0 ) {
        if ( d > 0.0 ) {
          if ( light.ls[i].position.w != 0.0 )
            d *= attFactor ( light.ls[i].attenuation, dist );
          Colour += (d * light.ls[i].direct) * mm.dirref;
        }
      }
      else {
        if ( d < 0.0 ) {
          if ( light.ls[i].position.w != 0.0 )
            d *= attFactor ( light.ls[i].attenuation, dist );
          Colour -= (d * light.ls[i].direct) * mm.dirref;
        }
      }
    }
  return clamp ( Colour, 0.0, 1.0 );
} /*LambertLighting*/

float wFactor ( float lvn, float wa, float we )
{
  return 1.0+(wa-1.0)*pow ( 1.0-lvn, we );
} /*wFactor*/

vec3 BlinnPhongLighting ( void )
{
  vec3  lv, vv, hv, Colour;
  float a, d, e, f, dist;
  uint  i, mask;

  vv = posDifference ( trb.eyepos, In.Position, dist );
  e = dot ( vv, tnormal );
  Colour = vec3(0.0);
  for ( i = 0, mask = 0x00000001;  i < light.nls;  i++, mask <<= 1 )
    if ( (light.mask & mask) != 0 ) {
      Colour += light.ls[i].ambient * mm.ambref;
      lv = posDifference ( light.ls[i].position, In.Position, dist );
      d = dot ( lv, normal );
      if ( e > 0.0 ) {
        if ( d > 0.0 ) {
          if ( light.ls[i].position.w != 0.0 )
            a = attFactor ( light.ls[i].attenuation, dist );
          else
            a = 1.0;
          Colour += (a * d * light.ls[i].direct) * mm.dirref;
          hv = normalize ( lv+vv );
          f = pow ( dot ( hv, normal ), mm.shininess );
          Colour += (a*f*wFactor(d,mm.wa,mm.we))*mm.specref;
        }
      }
      else {
        if ( d < 0.0 ) {
          if ( light.ls[i].position.w != 0.0 )
            a = attFactor ( light.ls[i].attenuation, dist );
          else
            a = 1.0;
          Colour -= (a * d * light.ls[i].direct) * mm.dirref;
          hv = normalize ( lv+vv );
          f = pow ( -dot ( hv, normal ), mm.shininess );
          Colour += (a*f*wFactor(-d,mm.wa,mm.we))*mm.specref;
        }
      }
    }
  return clamp ( Colour, 0.0, 1.0 );
} /*BlinnPhongLighting*/

#define AGamma(colour) pow ( colour, vec3(256.0/563.0) )

void main ( void )
{
  vec3 Colour;

  normal = normalize ( In.Normal );
  tnormal = In.TNormal;
  GetColours ();
  switch ( LightingModel ) {
default: Colour = LambertLighting ();     break;
 case 1: Colour = BlinnPhongLighting ();  break;
  }
  out_Colour = vec4 ( AGamma ( Colour ), 1.0 );
} /*main*/
