#version 450 core

#define MAX_DEG 6

layout(quads,equal_spacing,cw) in;

out GVertex {
    vec3 Colour;
    vec3 Position;
    vec3 Normal;
  } Out;

layout(std430,binding=0) buffer BSKnots { float uv[]; };
layout(std430,binding=1) buffer BSCPoints { float cp[]; };
layout(std430,binding=2) buffer BSPatch {
    int  dim, n, N, m, M, stride;
    vec3 Colour;
    bool BSPNormals;
    int  tesslevel;
  } bsp;

uniform TransBlock {
    mat4 mm, mmti, vm, pm, vpm;
    vec4 eyepos;
  } trb;

void BSCdeBoor2f ( int n, int k, vec2 d[MAX_DEG+1], float t, out vec2 p0 )
{
  int i, j;

  for ( j = 1; j <= n; j++ )
    for ( i = 0; i <= n-j; i++ )
      d[i] = mix ( d[i], d[i+1], (t-uv[k-n+i+j])/(uv[k+1+i]-uv[k-n+i+j]) );
  p0 = d[0];
} /*BSCdeBoor2f*/

void BSPdeBoor2f ( float u, int kk, float v, int ll,
                   out vec4 pos, out vec4 nv )
{
  vec2 p[MAX_DEG+1], q[MAX_DEG+1], p00;
  int  i, j, k, l;

  for ( i = 0, k = (kk-bsp.n)*bsp.stride;  i <= bsp.n;  i++, k += bsp.stride ) {
    for ( j = 0, l = k+2*(ll-bsp.N-1-bsp.m);  j <= bsp.m;  j++, l += 2 )
      p[j] = vec2 ( cp[l], cp[l+1] );
    BSCdeBoor2f ( bsp.m, ll, p, v, q[i] );
  }
  BSCdeBoor2f ( bsp.n, kk, q, u, p00 );
  pos = vec4 ( p00, 0.0, 1.0 );
  nv = vec4 ( 0.0, 0.0, 1.0, 0.0 );
} /*BSPdeBoor2f*/

void BSCdeBoor3f ( int n, int k, vec3 d[MAX_DEG+1], float t,
                   out vec3 p0, out vec3 p1, out float a )
{
  int i, j;

  for ( j = 1; j < n; j++ )
    for ( i = 0; i <= n-j; i++ )
      d[i] = mix ( d[i], d[i+1], (t-uv[k-n+i+j])/(uv[k+1+i]-uv[k-n+i+j]) );
  a = (t-uv[k])/(uv[k+1]-uv[k]);
  p0 = d[0];  p1 = d[1];
} /*BSCdeBoor3f*/

void BSPdeBoor3f ( float u, int kk, float v, int ll,
                   out vec4 pos, out vec4 nv )
{
  vec3  p[MAX_DEG+1], q0[MAX_DEG+1], q1[MAX_DEG+1],
        p00, p01, p10, p11, a, b, ru, rv;
  float au, av;
  int   i, j, k, l;

  for ( i = 0, k = (kk-bsp.n)*bsp.stride;  i <= bsp.n;  i++, k += bsp.stride ) {
    for ( j = 0, l = k+3*(ll-bsp.N-1-bsp.m);  j <= bsp.m;  j++, l += 3 )
      p[j] = vec3 ( cp[l], cp[l+1], cp[l+2] );
    BSCdeBoor3f ( bsp.m, ll, p, v, q0[i], q1[i], av );
  }
  BSCdeBoor3f ( bsp.n, kk, q0, u, p00, p10, au );
  BSCdeBoor3f ( bsp.n, kk, q1, u, p01, p11, au );
  ru = (b = mix ( p10, p11, av )) - (a = mix ( p00, p01, av ));
  rv = mix ( p01-p00, p11-p10, au );
  pos = vec4 ( mix ( a, b, au ), 1.0 );
  nv = vec4 ( cross ( ru, rv ), 0.0 );
} /*BSPdeBoor3f*/

vec4 cross4 ( vec4 v0, vec4 v1, vec4 v2 )
{
  float a01, a02, a03, a12, a13, a23;

  a01 = v0.x*v1.y - v0.y*v1.x;  a02 = v0.x*v1.z - v0.z*v1.x;
  a03 = v0.x*v1.w - v0.w*v1.x;  a12 = v0.y*v1.z - v0.z*v1.y;
  a13 = v0.y*v1.w - v0.w*v1.y;  a23 = v0.z*v1.w - v0.w*v1.z;
  return vec4 ( -a23*v2.y+a13*v2.z-a12*v2.w, a23*v2.x-a03*v2.z+a02*v2.w,
                -a13*v2.x+a03*v2.y-a01*v2.w, a12*v2.x-a02*v2.y+a01*v2.z );
} /*cross4*/

void BSCdeBoor4f ( int n, int k, vec4 d[MAX_DEG+1], float t,
                   out vec4 p0, out vec4 p1, out float a )
{
  int i, j;

  for ( j = 1; j < n; j++ )
    for ( i = 0; i <= n-j; i++ )
      d[i] = mix ( d[i], d[i+1], (t-uv[k-n+i+j])/(uv[k+1+i]-uv[k-n+i+j]) );
  a = (t-uv[k])/(uv[k+1]-uv[k]);
  p0 = d[0];  p1 = d[1];
} /*BSCdeBoor4f*/

void BSPdeBoor4f ( float u, int kk, float v, int ll,
                   out vec4 pos, out vec4 nv )
{
  vec4  p[MAX_DEG+1], q0[MAX_DEG+1], q1[MAX_DEG+1],
        p00, p01, p10, p11, a, b, ru, rv;
  float au, av;
  int   i, j, k, l;

  for ( i = 0, k = (kk-bsp.n)*bsp.stride;  i <= bsp.n;  i++, k += bsp.stride ) {
    for ( j = 0, l = k+3*(ll-bsp.N-1-bsp.m); j <= bsp.m; j++, l += 4 )
      p[j] = vec4 ( cp[l], cp[l+1], cp[l+2], cp[l+4] );
    BSCdeBoor4f ( bsp.m, ll, p, v, q0[i], q1[i], av );
  }
  BSCdeBoor4f ( bsp.n, kk, q0, u, p00, p10, au );
  BSCdeBoor4f ( bsp.n, kk, q1, u, p01, p11, au );
  ru = (b = mix ( p10, p11, av )) - (a = mix ( p00, p01, av ));
  rv = mix ( p01-p00, p11-p10, au );
  pos = mix ( a, b, au );
  nv = cross4 ( pos, rv, ru );
} /*BSPdeBoor4f*/

int FindKnotInterval ( int k0, int kN, int n, float t )
{
  int i, j, k;

  for ( i = k0+n, j = kN-n;  j-i > 1; ) {
    k = i + (j-i)/2;
    if ( t >= uv[k] ) i = k; else j = k;
  }
  return i;
} /*FindKnotInterval*/

void main ( void )
{
  float u, v;
  int   k, l, m0, m1;
  vec4  pos, nv;

  u = uv[bsp.n] + gl_TessCoord.x*(uv[bsp.N-bsp.n] - uv[bsp.n]);
  m0 = bsp.N+1;  m1 = bsp.N+1+bsp.M;
  v = uv[m0+bsp.m] + gl_TessCoord.y*(uv[m1-bsp.m] - uv[m0+bsp.m]);
  k = FindKnotInterval ( 0, bsp.N, bsp.n, u );
  l = FindKnotInterval ( m0, m1, bsp.m, v );
  switch ( bsp.dim ) {
 case 2: BSPdeBoor2f ( u, k, v, l, pos, nv );  break;
 case 3: BSPdeBoor3f ( u, k, v, l, pos, nv );  break;
 case 4: BSPdeBoor4f ( u, k, v, l, pos, nv );  break;
default: pos = nv = vec4 ( 0.0 );
  }
  pos = trb.mm * pos;
  gl_Position = trb.vpm * pos;
  Out.Position = pos.xyz;
  if ( !bsp.BSPNormals || dot ( nv, nv ) < 1.0e-10 )
    Out.Normal = vec3 ( 0.0 );
  else
    Out.Normal = normalize ( (trb.mmti*nv).xyz );
  Out.Colour = bsp.Colour;
} /*main*/
