
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "openglheader.h"
#include <GLFW/glfw3.h>

#include "utilities.h"
#include "GPUsparsemat.h"

typedef struct {
    GPUSparseMatrix mat;
    GLuint          *rows, *cols, *aux;
    GLfloat         *coeff;
  } matdesc;

GLFWwindow *mywindow;
int win_width, win_height;

char redraw = 0;

void myGLFWErrorHandler ( int error, const char *description )
{
  fprintf ( stderr, "GLFW error: %s\n", description );
  exit ( 1 );
} /*myGLFWErrorHandler*/

/* ////////////////////////////////////////////////////////////////////////// */
char Less ( void *data, int i, int j )
{
  matdesc *aux;
  int i2, j2;

  aux = (matdesc*)data;
  i2 = i+i;  j2 = j+j;
  if ( aux->aux[i2] < aux->aux[j2] )
    return true;
  else if ( aux->aux[i2] > aux->aux[j2] )
    return false;
  else
    return aux->aux[i2+1] < aux->aux[j2+1];
} /*Less*/

void Swap ( void *data, int i, int j )
{
  matdesc *aux;
  int     i2, j2, temp;
  GLfloat ftemp;

  aux = (matdesc*)data;
  i2 = i+i;  j2 = j+j;
  temp = aux->aux[i2];    aux->aux[i2] = aux->aux[j2];      aux->aux[j2] = temp;
  temp = aux->aux[i2+1];  aux->aux[i2+1] = aux->aux[j2+1];  aux->aux[j2+1] = temp;
  ftemp = aux->coeff[i];  aux->coeff[i] = aux->coeff[j];    aux->coeff[j] = ftemp;
} /*Swap*/

void ReadTheMatrix ( const char *fn, matdesc *mat )
{
  FILE    *f;
  int     i, j, k;
  matdesc aux;

  if ( (f = fopen ( fn, "r")) ) {
    fscanf ( f, "%d %d %d", &mat->mat.m, &mat->mat.n, &mat->mat.nnz );
    printf ( "m = %d, n = %d, nnz = %d\n",
             mat->mat.m, mat->mat.n, mat->mat.nnz );
    aux.mat.m = mat->mat.m;  aux.mat.n = mat->mat.n;
    aux.mat.nnz = mat->mat.nnz;
    mat->rows = aux.rows = (unsigned int*)malloc ( (mat->mat.m+1)*sizeof(int) );
    mat->cols = aux.cols = (unsigned int*)malloc ( mat->mat.nnz*sizeof(int) );
    aux.aux = (unsigned int*)malloc ( 2*mat->mat.nnz*sizeof(int) );
    aux.coeff = mat->coeff = (float*)malloc ( mat->mat.nnz*sizeof(float) );
    for ( i = 0; i < mat->mat.nnz; i++ )
      fscanf ( f, "%u %u %f", &aux.aux[2*i], &aux.aux[2*i+1], &mat->coeff[i] );
    fclose ( f );
    QuickSort ( &aux, mat->mat.nnz, Less, Swap );
    mat->rows[0] = 0;
    for ( i = 1, j = 0; j < mat->mat.nnz; j++ ) {
      mat->cols[j] = aux.aux[2*j+1];
      if ( aux.aux[2*j] >= i )
        mat->rows[i++] = j;
    }
    mat->rows[mat->mat.m] = mat->mat.nnz;
    free ( aux.aux );
    j = k = mat->rows[1];
    for ( i = 1; i < mat->mat.m; i++ )
      if ( mat->rows[i+1]-mat->rows[i] > j )
        j =  mat->rows[i+1]-mat->rows[i];
      else if ( mat->rows[i+1]-mat->rows[i] < k )
        k = mat->rows[i+1]-mat->rows[i];
    printf ( "row min = %d, row max = %d\n", k, j );
    glGenBuffers ( 2, mat->mat.buf );
    glBindBuffer ( GL_SHADER_STORAGE_BUFFER, mat->mat.buf[0] );
    glBufferData ( GL_SHADER_STORAGE_BUFFER,
        (mat->mat.m+1+mat->mat.nnz)*sizeof(GLuint), NULL, GL_STATIC_DRAW );
    glBufferSubData ( GL_SHADER_STORAGE_BUFFER, 0,
        (mat->mat.m+1)*sizeof(GLuint), mat->rows );
    glBufferSubData ( GL_SHADER_STORAGE_BUFFER,
        (mat->mat.m+1)*sizeof(GLuint), mat->mat.nnz*sizeof(GLuint), mat->cols );
    glBindBuffer ( GL_SHADER_STORAGE_BUFFER, mat->mat.buf[1] );
    glBufferData ( GL_SHADER_STORAGE_BUFFER, mat->mat.nnz*sizeof(GLfloat),
                   mat->coeff, GL_STATIC_DRAW );
    ExitIfGLError ( "ReadTheMatrix" );
  }
  else
    ExitOnError ( "can't read" );
} /*ReadTheMatrix*/

/* ////////////////////////////////////////////////////////////////////////// */
void TestSparseMM1 ( void )
{
/*
#define M 6
#define N 3
#define L 4
#define NNZA 10
#define NNZB 7
  static GLuint  rca[] = {0,2,4,5,7,8,10, 0,2,0,1,1,1,2,0,0,1};
  static GLfloat ca[]  = {1.0,2.0,3.0,4.0,5.0,6.0,7.0,8.0,9.0,10.0};
  static GLuint  rcb[] = {0,2,5,7, 0,3,0,1,2,1,3};
  static GLfloat cb[]  = {1.0,2.0,3.0,4.0,5.0,6.0,7.0};
*/
#define M 7
#define N 3
#define L 4
#define NNZA 10
#define NNZB 7
  static GLuint  rca[] = {0,2,4,5,7,8,8,10, 0,2,0,1,1,1,2,0,0,1};
  static GLfloat ca[]  = {1.0,2.0,3.0,4.0,5.0,6.0,7.0,8.0,9.0,10.0};
  static GLuint  rcb[] = {0,2,5,7, 0,3,0,1,2,1,3};
  static GLfloat cb[]  = {1.0,2.0,3.0,4.0,5.0,6.0,7.0};
  GPUSparseMatrix a, b, c;
  GLuint  mb[6];
  GLuint  *rcc;
  GLfloat *ac;
  int     i;

  printf ( "Test 1\n" );
  glGenBuffers ( 4, mb );
  a.buf[0] = mb[0];  a.buf[1] = mb[1];  a.m = M;  a.n = N;  a.nnz = NNZA;  a.lmax = 0;
  b.buf[0] = mb[2];  b.buf[3] = mb[1];  b.m = N;  b.n = L;  b.nnz = NNZB;  b.lmax = 0;
  glBindBuffer ( GL_SHADER_STORAGE_BUFFER, mb[0] );
  glBufferData ( GL_SHADER_STORAGE_BUFFER, (M+1+NNZA)*sizeof(GLuint), rca, GL_STATIC_DRAW );
  glBindBuffer ( GL_SHADER_STORAGE_BUFFER, mb[1] );
  glBufferData ( GL_SHADER_STORAGE_BUFFER, NNZA*sizeof(GLfloat), ca, GL_STATIC_DRAW );
  glBindBuffer ( GL_SHADER_STORAGE_BUFFER, mb[2] );
  glBufferData ( GL_SHADER_STORAGE_BUFFER, (N+1+NNZB)*sizeof(GLuint), rcb, GL_STATIC_DRAW );
  glBindBuffer ( GL_SHADER_STORAGE_BUFFER, mb[3] );
  glBufferData ( GL_SHADER_STORAGE_BUFFER, NNZB*sizeof(GLfloat), cb, GL_STATIC_DRAW );
  if ( GPUMultSparseMatricesf ( &c, &a, &b ) ) {
    if ( c.nnz > 0 ) {
      rcc = malloc ( (M+1+c.nnz)*sizeof(GLuint) );
      ac = malloc ( c.nnz*sizeof(GLfloat) );
      glBindBuffer ( GL_SHADER_STORAGE_BUFFER, c.buf[0] );
      glGetBufferSubData ( GL_SHADER_STORAGE_BUFFER,
                           0, (M+1+c.nnz)*sizeof(GLuint), rcc );
      glBindBuffer ( GL_SHADER_STORAGE_BUFFER, c.buf[1] );
      glGetBufferSubData ( GL_SHADER_STORAGE_BUFFER,
                           0, c.nnz*sizeof(GLuint), ac );
      printf ( "nnzc = %d\n", c.nnz );
      for ( i = 0; i <= M; i++ )
        printf ( "%2d,", rcc[i] );
      printf ( "\n" );
      for ( ; i <= M+c.nnz; i++ )
        printf ( "%5d,", rcc[i] );
      printf ( "\n" );
      for ( i = 0; i < c.nnz; i++ )
        printf ( "%5.2f,", ac[i] );
      printf ( "\n" );
    }
    printf ( "OK!\n" );
  }
} /*TestSparseMM1*/

void TestSparseMM2 ( void )
{
  FILE    *f;
  matdesc a, b, c;
  GLuint  *t1;
  GLfloat *t2;
  GLuint  i, j;

  printf ( "Test 2\n" );
  ReadTheMatrix ( "spm.txt", &a );
  ReadTheMatrix ( "mat1.txt", &b );
  if ( GPUMultSparseMatricesf ( &c.mat, &a.mat, &b.mat ) ) {
    t1 = malloc ( (c.mat.m+1+c.mat.nnz)*sizeof(GLuint) );
    t2 = malloc ( c.mat.nnz*sizeof(GLfloat) );
    glBindBuffer ( GL_SHADER_STORAGE_BUFFER, c.mat.buf[0] );
    glGetBufferSubData ( GL_SHADER_STORAGE_BUFFER, 0,
                         (c.mat.m+1+c.mat.nnz)*sizeof(GLuint), t1 );
    glBindBuffer ( GL_SHADER_STORAGE_BUFFER, c.mat.buf[1] );
    glGetBufferSubData ( GL_SHADER_STORAGE_BUFFER, 0,
                         c.mat.nnz*sizeof(GLfloat), t2 );
    f = fopen ( "matprod.txt", "w" );
    fprintf ( f, "%d %d %d\n", c.mat.m, c.mat.n, c.mat.nnz );
    for ( i = 0; i < c.mat.m; i++ ) {
      for ( j = t1[i]; j < t1[i+1]; j++ )
        fprintf ( f, "%d %d %8.5f\n", i, t1[c.mat.m+1+j], t2[j] );
    }
    fclose ( f );
    free ( t1 );
    free ( t2 );
    glDeleteBuffers ( 2, a.mat.buf );
    glDeleteBuffers ( 2, b.mat.buf );
    glDeleteBuffers ( 2, c.mat.buf );
    printf ( "OK!\n" );
  }
} /*TestSparseMM2*/

/* ////////////////////////////////////////////////////////////////////////// */
void InitMyObject ( void )
{
  LoadGPUMultMMfShader ();
  TestSparseMM1 ();
  TestSparseMM2 ();
  ExitIfGLError ( "InitMyObject" );
} /*InitMyObject*/

void ReshapeFunc ( GLFWwindow *win, int width, int height )
{
} /*ReshapeFunc*/

void Redraw ( GLFWwindow *win )
{
  glClearColor ( 1.0, 1.0, 1.0, 1.0 );
  glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
  glUseProgram ( 0 );
  glFlush ();
  ExitIfGLError ( "Redraw" );
  redraw = false;
  glfwSwapBuffers ( win );
} /*Redraw*/

void DisplayFunc ( GLFWwindow *win )
{
  redraw = true; 
} /*DisplayFunc*/

void KeyFunc ( GLFWwindow *win, int key, int scancode, int action, int mode )
{
  if ( action == GLFW_PRESS || action == GLFW_REPEAT ) {
    switch ( key ) {   
  case GLFW_KEY_ESCAPE:
      glfwSetWindowShouldClose ( mywindow, 1 );
      break;
  default:  
      break;
    }
  }
} /*KeyFunc*/

void CharFunc ( GLFWwindow *win, unsigned int charcode )
{
} /*CharFunc*/

void MouseFunc ( GLFWwindow *win, int button, int action, int mods )
{
} /*MouseFunc*/

void MotionFunc ( GLFWwindow *win, double x, double y )
{
} /*MotionFunc*/

void ScrollFunc ( GLFWwindow *win, double x, double y )
{
} /*ScrollFunc*/

void IdleFunc ( void )
{
} /*IdleFunc*/

void Initialise ( int argc, char **argv )
{
  glfwSetErrorCallback ( myGLFWErrorHandler );
  if ( !glfwInit () )
    ExitOnError ( "Initialise 0" );
  if ( !(mywindow = glfwCreateWindow ( 480, 360,
                                       "Sortnet", NULL, NULL )) ) {
    glfwTerminate ();
    ExitOnError ( "Initialise 1" );
  }
  glfwMakeContextCurrent ( mywindow );
  GetGLProcAddresses ( 4, 3 );
  PrintGLVersion ();
  glfwSetFramebufferSizeCallback ( mywindow, ReshapeFunc );
  glfwSetWindowRefreshCallback ( mywindow, DisplayFunc );
  glfwSetKeyCallback ( mywindow, KeyFunc );  
  glfwSetCharCallback ( mywindow, CharFunc );
  glfwSetMouseButtonCallback ( mywindow, MouseFunc );
  glfwSetCursorPosCallback ( mywindow, MotionFunc );
  glfwSetScrollCallback ( mywindow, ScrollFunc );
  InitMyObject ();
  ReshapeFunc ( mywindow, 480, 360 );
  redraw = true;
  ExitIfGLError ( "Initialise" );
} /*Initialise*/

void MessageLoop ( void )
{
  do {
    glfwWaitEvents ();
    if ( redraw )
      Redraw ( mywindow );
  } while ( !glfwWindowShouldClose ( mywindow ) );
} /*MessageLoop*/

void Cleanup ( void )
{
} /*Cleanup*/

int main ( int argc, char **argv )
{
  Initialise ( argc, argv );
  MessageLoop ();
  Cleanup ();
  exit ( 0 );
} /*main*/

