
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include "openglheader.h"
#include <GLFW/glfw3.h>

#include "utilities.h"
#include "GPUsparsemat.h"

typedef struct {
    GPUSparseMatrix mat;
    GLuint          *rows, *cols, *aux;
    GLfloat         *coeff;
  } matdesc;

GLFWwindow *mywindow;
int win_width, win_height;

char redraw = 0;

void myGLFWErrorHandler ( int error, const char *description )
{
  fprintf ( stderr, "GLFW error: %s\n", description );
  exit ( 1 );
} /*myGLFWErrorHandler*/

/* ////////////////////////////////////////////////////////////////////////// */
char Less ( void *data, int i, int j )
{
  matdesc *aux;
  int i2, j2;

  aux = (matdesc*)data;
  i2 = i+i;  j2 = j+j;
  if ( aux->aux[i2] < aux->aux[j2] )
    return true;
  else if ( aux->aux[i2] > aux->aux[j2] )
    return false;
  else
    return aux->aux[i2+1] < aux->aux[j2+1];
} /*Less*/

void Swap ( void *data, int i, int j )
{
  matdesc *aux;
  int     i2, j2, temp;
  GLfloat ftemp;

  aux = (matdesc*)data;
  i2 = i+i;  j2 = j+j;
  temp = aux->aux[i2];    aux->aux[i2] = aux->aux[j2];      aux->aux[j2] = temp;
  temp = aux->aux[i2+1];  aux->aux[i2+1] = aux->aux[j2+1];  aux->aux[j2+1] = temp;
  ftemp = aux->coeff[i];  aux->coeff[i] = aux->coeff[j];    aux->coeff[j] = ftemp;
} /*Swap*/

void ReadTheMatrix ( const char *fn, matdesc *mat )
{
  FILE    *f;
  int     i, j, k;
  matdesc aux;

  if ( (f = fopen ( fn, "r")) ) {
    fscanf ( f, "%d %d %d", &mat->mat.m, &mat->mat.n, &mat->mat.nnz );
    printf ( "m = %d, n = %d, nnz = %d\n",
             mat->mat.m, mat->mat.n, mat->mat.nnz );
    aux.mat.m = mat->mat.m;  aux.mat.n = mat->mat.n;
    aux.mat.nnz = mat->mat.nnz;
    mat->rows = aux.rows = (unsigned int*)malloc ( (mat->mat.m+1)*sizeof(GLuint) );
    mat->cols = aux.cols = (unsigned int*)malloc ( mat->mat.nnz*sizeof(GLuint) );
    aux.aux = (unsigned int*)malloc ( 2*mat->mat.nnz*sizeof(GLuint) );
    aux.coeff = mat->coeff = (float*)malloc ( mat->mat.nnz*sizeof(float) );
    for ( i = 0; i < mat->mat.nnz; i++ )
      fscanf ( f, "%u %u %f", &aux.aux[2*i], &aux.aux[2*i+1], &mat->coeff[i] );
    fclose ( f );
    QuickSort ( &aux, mat->mat.nnz, Less, Swap );
    mat->rows[0] = 0;
    for ( i = 1, j = 0; j < mat->mat.nnz; j++ ) {
      mat->cols[j] = aux.aux[2*j+1];
      if ( aux.aux[2*j] >= i )
        mat->rows[i++] = j;
    }
    mat->rows[mat->mat.m] = mat->mat.nnz;
    free ( aux.aux );
    j = k = mat->rows[1];
    for ( i = 1; i < mat->mat.m; i++ )
      if ( mat->rows[i+1]-mat->rows[i] > j )
        j =  mat->rows[i+1]-mat->rows[i];
      else if ( mat->rows[i+1]-mat->rows[i] < k )
        k = mat->rows[i+1]-mat->rows[i];
    printf ( "row min = %d, row max = %d\n", k, j );
    glGenBuffers ( 2, mat->mat.buf );
    glBindBuffer ( GL_SHADER_STORAGE_BUFFER, mat->mat.buf[0] );
    glBufferData ( GL_SHADER_STORAGE_BUFFER,
        (mat->mat.m+1+mat->mat.nnz)*sizeof(GLuint), NULL, GL_STATIC_DRAW );
    glBufferSubData ( GL_SHADER_STORAGE_BUFFER, 0,
        (mat->mat.m+1)*sizeof(GLuint), mat->rows );
    glBufferSubData ( GL_SHADER_STORAGE_BUFFER,
        (mat->mat.m+1)*sizeof(GLuint), mat->mat.nnz*sizeof(GLuint), mat->cols );
    glBindBuffer ( GL_SHADER_STORAGE_BUFFER, mat->mat.buf[1] );
    glBufferData ( GL_SHADER_STORAGE_BUFFER, mat->mat.nnz*sizeof(GLfloat),
                   mat->coeff, GL_STATIC_DRAW );
    ExitIfGLError ( "ReadTheMatrix" );
  }
  else
    ExitOnError ( "can't read" );
} /*ReadTheMatrix*/

void CheckIt ( int m, unsigned int *rows, unsigned int *cols, GLfloat *coeff,
               GLfloat *x, GLfloat *y )
{
  int i, j;

  memset ( y, 0, m*sizeof(GLfloat) );
  for ( i = 0; i < m; i++ )
    for ( j = rows[i]; j < rows[i+1]; j++ )
      y[i] += coeff[j]*x[cols[j]];
} /*CheckIt*/

/* ////////////////////////////////////////////////////////////////////////// */
void InitMyObject ( void )
{
  matdesc mat;
  GLfloat *x, *y, *z;
  GLuint  xybuf[2];
  int     i;

  LoadGPUMultMVfShader ();
  ReadTheMatrix ( "spm.txt", &mat );
  glGenBuffers ( 2, xybuf );
  x = malloc ( mat.mat.n*sizeof(GLfloat) );
  y = malloc ( 2*mat.mat.m*sizeof(GLfloat) );
  z = &y[mat.mat.m];
  for ( i = 0; i < mat.mat.n; i++ )
    x[i] = 1.0;
  CheckIt ( mat.mat.m, mat.rows, mat.cols, mat.coeff, x, z );
  memset ( y, 0, mat.mat.m*sizeof(GLfloat) );
  glBindBuffer ( GL_SHADER_STORAGE_BUFFER, xybuf[0] ); 
  glBufferData ( GL_SHADER_STORAGE_BUFFER, mat.mat.n*sizeof(GLfloat), x, GL_STATIC_DRAW );
  glBindBuffer ( GL_SHADER_STORAGE_BUFFER, xybuf[1] );
  glBufferData ( GL_SHADER_STORAGE_BUFFER, mat.mat.m*sizeof(GLfloat), NULL, GL_DYNAMIC_DRAW );
  mat.mat.lmax = 0;
  GPUMultSparseMatrixVectorf ( xybuf[1], &mat.mat, 1, xybuf[0] );
  glBindBuffer ( GL_SHADER_STORAGE_BUFFER, xybuf[1] );
  glGetBufferSubData ( GL_SHADER_STORAGE_BUFFER, 0, mat.mat.m*sizeof(GLfloat), y );
  for ( i = 0; i < mat.mat.m; i++ )
    printf ( "%5.2f,", y[i] );
  printf ( "\n" );
  free ( x );
  free ( y );
  ExitIfGLError ( "InitMyObject" );
} /*InitMyObject*/

void ReshapeFunc ( GLFWwindow *win, int width, int height )
{
} /*ReshapeFunc*/

void Redraw ( GLFWwindow *win )
{
  glClearColor ( 1.0, 1.0, 1.0, 1.0 );
  glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
  glUseProgram ( 0 );
  glFlush ();
  ExitIfGLError ( "Redraw" );
  redraw = false;
  glfwSwapBuffers ( win );
} /*Redraw*/

void DisplayFunc ( GLFWwindow *win )
{
  redraw = true; 
} /*DisplayFunc*/

void KeyFunc ( GLFWwindow *win, int key, int scancode, int action, int mode )
{
  if ( action == GLFW_PRESS || action == GLFW_REPEAT ) {
    switch ( key ) {   
  case GLFW_KEY_ESCAPE:
      glfwSetWindowShouldClose ( mywindow, 1 );
      break;
  default:  
      break;
    }
  }
} /*KeyFunc*/

void CharFunc ( GLFWwindow *win, unsigned int charcode )
{
} /*CharFunc*/

void MouseFunc ( GLFWwindow *win, int button, int action, int mods )
{
} /*MouseFunc*/

void MotionFunc ( GLFWwindow *win, double x, double y )
{
} /*MotionFunc*/

void ScrollFunc ( GLFWwindow *win, double x, double y )
{
} /*ScrollFunc*/

void IdleFunc ( void )
{
} /*IdleFunc*/

void Initialise ( int argc, char **argv )
{
  glfwSetErrorCallback ( myGLFWErrorHandler );
  if ( !glfwInit () )
    ExitOnError ( "Initialise 0" );
  if ( !(mywindow = glfwCreateWindow ( 480, 360,
                                       "Sortnet", NULL, NULL )) ) {
    glfwTerminate ();
    ExitOnError ( "Initialise 1" );
  }
  glfwMakeContextCurrent ( mywindow );
  GetGLProcAddresses ( 4, 3 );
  PrintGLVersion ();
  glfwSetFramebufferSizeCallback ( mywindow, ReshapeFunc );
  glfwSetWindowRefreshCallback ( mywindow, DisplayFunc );
  glfwSetKeyCallback ( mywindow, KeyFunc );  
  glfwSetCharCallback ( mywindow, CharFunc );
  glfwSetMouseButtonCallback ( mywindow, MouseFunc );
  glfwSetCursorPosCallback ( mywindow, MotionFunc );
  glfwSetScrollCallback ( mywindow, ScrollFunc );
  InitMyObject ();
  ReshapeFunc ( mywindow, 480, 360 );
  redraw = true;
  ExitIfGLError ( "Initialise" );
} /*Initialise*/

void MessageLoop ( void )
{
  do {
    glfwWaitEvents ();
    if ( redraw )
      Redraw ( mywindow );
  } while ( !glfwWindowShouldClose ( mywindow ) );
} /*MessageLoop*/

void Cleanup ( void )
{
} /*Cleanup*/

int main ( int argc, char **argv )
{
  Initialise ( argc, argv );
  MessageLoop ();
  Cleanup ();
  exit ( 0 );
} /*main*/

