#version 440 core

#define INF0 65533
#define INF1 65534
#define INF2 65535

#define MAXDEG      10

uniform TransBlock {
    dvec4 rxy;
    dvec2 wh;
    int   width, mag, maxit, currit, diter;
  } tr;

uniform PaletteBlock {
    int  deg;
    vec4 cp[MAXDEG+1], bpcp[MAXDEG+1];
  } pal;

uniform int pn = 1;

layout(location=0) out vec4 out_Colour;

layout(r32f,binding=1) uniform image2D img;

vec3 Palette0 ( uint itr )
{
  switch ( itr ) {
case INF0: return vec3 ( 0.0, 0.0, 1.0 );
case INF1: return vec3 ( 0.0, 1.0, 0.0 );
case INF2: return vec3 ( 1.0, 0.0, 0.0 );
  default: return (itr & 0x01) != 0 ? vec3 ( 1.0, 1.0, 1.0 ) :
                  vec3 ( 0.0, 0.0, 0.0 );
  }
} /*Palette0*/

vec3 BCHorner3Rf ( int n, vec4 bcp[MAXDEG+1], float t )
{
  int   i, b;
  float s, d;
  vec4  p;

  s = 1.0-t;  d = t;  b = n;
  p = bcp[0];
  for ( i = 1; i <= n; i++ ) {
    p = s*p + (b*d)*bcp[i];
    d *= t;  b = (b*(n-i))/(i+1);
  }
  return p.xyz/p.w;
} /*BCHorner3Rf*/

vec3 Palette1 ( float itr )
{
  float t;
  vec3  p, c;

  if ( itr >= INF0 )
    return vec3 ( 0.0, 0.0, 0.0 );
  else {
    t = itr / float(tr.maxit);
    p = BCHorner3Rf ( pal.deg, pal.cp, t );
    c.rg = p.xy;  c.b = 1.0 - p.x - p.y;
    c = clamp ( c, vec3(0.0), vec3(1.0) );
    t = max ( c.r, c.g );  t = max ( t, c.b );
    return p.z*c/t;
  }
} /*Palette1*/

vec3 Palette ( float tx )
{
  switch ( pn ) {
 default: return Palette0 ( uint(tx) );
  case 1: return Palette1 ( tx );
  case 2: return Palette1 ( trunc(tx) );
  }
} /*Palette*/

ivec2 wh;

float itx ( ivec2 xy )
{
  if ( xy.x < 0 || xy.y < 0 || xy.x >= wh.x || xy.y >= wh.y )
    return float(INF0);
  else
    return imageLoad ( img, xy ).r;
} /*itx*/

#define AGamma(colour) pow ( colour, vec3(256.0/563.0) )

void main ( void )
{
  const float mgf[4] =
               {0.3831031644, 0.2418428568, 0.0606257426, 0.0059798184};
  int   i, j;
  ivec2 xy;
  float tx;
  vec3  Colour, col;

  wh = imageSize ( img );
  xy = ivec2 ( int(gl_FragCoord.x), int(gl_FragCoord.y) );
  if ( tr.mag == 1 )
    Colour = Palette ( uint(itx ( xy)) );
  else {  /* mag == 3 */
    xy = tr.mag * xy - ivec2 (-1,-1);
    col = mgf[0] * Palette ( itx ( xy ) );
    for ( j = 1; j < 4; j++ )
      col += mgf[j] * (Palette ( itx ( xy+ivec2(0, j) ) ) +
                       Palette ( itx ( xy+ivec2(0,-j) ) ));
    Colour = mgf[0] * col;
    for ( i = 1; i < 4;  i++ ) {
      col = mgf[0] * (Palette ( itx ( xy+ivec2( i,0) ) ) +
                      Palette ( itx ( xy+ivec2(-i,0) ) ));
      for ( j = 1; j < 4; j++ )
        col += mgf[j] * (Palette ( itx ( xy+ivec2( i, j) ) ) +
                         Palette ( itx ( xy+ivec2( i,-j) ) ) +
                         Palette ( itx ( xy+ivec2(-i, j) ) ) +
                         Palette ( itx ( xy+ivec2(-i,-j) ) ));
      Colour += mgf[i] * col;
    }
  }
  out_Colour = vec4 ( AGamma ( Colour ), 1.0 );
} /*main*/

