#include <stdlib.h>
#include <stdio.h>
#include "openglheader.h"
#include <GLFW/glfw3.h>

#include "utilities.h"
#include "mygltext.h"
#include "mandelbrot.h"

GLFWwindow *window1, *window2;
double     last_xi1, last_eta1, last_xi2, last_eta2;
char       antialias1 = true, showtext1 = false;
char       redraw1, redraw2, queue_was_empty, opti1, opti2;
char       app_state = STATE_NOTHING;

void myGLFWErrorHandler ( int error, const char *description )
{
  fprintf ( stderr, "GLFW error: %s\n", description );
  exit ( 1 );
} /*myGLFWErrorHandler*/

void ReshapeFunc1 ( GLFWwindow *win, int width, int height )
{
  glfwMakeContextCurrent ( win );
  SetupTextFrame ( width, height );
  FindTheMapping ( width, height );
  opti1 = 4;
  redraw1 = true;
} /*ReshapeFunc1*/

void DisplayFunc1 ( GLFWwindow *win )
{
  redraw1 = true;
} /*DisplayFunc1*/

void Redraw1 ( GLFWwindow *win )
{
  glfwMakeContextCurrent ( win );
  if ( opti1 == 0 ) {
    if ( (redraw1 = !RedrawMyObject1 ()) )
      glfwPostEmptyEvent ();
  }
  else {
    opti1 --;
    glfwPostEmptyEvent ();
  }
  glfwSwapBuffers ( win );
  ExitIfGLError ( "Redraw" );
} /*Redraw1*/

void KeyboardFunc1 ( GLFWwindow *win, int key, int scancode, int action, int mods )
{
  queue_was_empty = false;
  if ( action == GLFW_PRESS || action == GLFW_REPEAT ) {
    switch ( key ) {
  case GLFW_KEY_ESCAPE:
      glfwSetWindowShouldClose ( window1, 1 );
      glfwSetWindowShouldClose ( window2, 1 );
      break;
  default:
      break;
    }
  }
} /*KeyboardFunc1*/

void CharFunc1 ( GLFWwindow *win, unsigned int charcode )
{
  queue_was_empty = false;
  switch ( charcode ) {
case '+':
    redraw1 |= ChangeMaxIter ( true );
    break;
case '-':
    redraw1 |= ChangeMaxIter ( false );
    break;
case 'A':  case 'a':
    antialias1 = !antialias1;
    FindTheMapping ( wdt, hgh );
    redraw1 = true;
    break;
case 'R':  case 'r':
    ResetTheMapping ();
    redraw1 = true;
    break;
case 'T':  case 't':
    showtext1 = !showtext1;
    redraw1 = true;
    break;
case '0':
    ChoosePalette ( 0 );
    redraw1 = true;
    break;
case '1':
    ChoosePalette ( 1 );
    redraw1 = true;
    break;
case '2':
    ChoosePalette ( 2 );
    redraw1 = true;
    break;
default:
    break;
  }
} /*CharFunc1*/

void ScrollFunc1 ( GLFWwindow *win, double x, double y )
{
  queue_was_empty = false;
  switch ( app_state ) {
case STATE_NOTHING:
    redraw1 |= Zoom ( y < 0.0 );
    break;
default:
    redraw1 |= ChangeMaxIter ( y < 0.0 );
    break;
  }
} /*ScrollFunc1*/

void MouseFunc1 ( GLFWwindow *win, int button, int action, int mods )
{
  queue_was_empty = false;
  if ( button == GLFW_MOUSE_BUTTON_LEFT ) {
    if ( action == GLFW_PRESS ) {
      app_state = STATE_PANNING1;
      glfwGetCursorPos ( win, &last_xi1, &last_eta1 );
    }
    else
      app_state = STATE_NOTHING;
  }
} /*MouseFunc1*/

void MotionFunc1 ( GLFWwindow *win, double x, double y )
{
  queue_was_empty = false;
  if ( app_state == STATE_PANNING1 ) {
    Pan ( x-last_xi1, y-last_eta1 );
    last_xi1 = x;  last_eta1 = y;
    redraw1 = true;
  }
} /*MotionFunc1*/

void ReshapeFunc2 ( GLFWwindow *win, int width, int height )
{
  queue_was_empty = false;
  glfwMakeContextCurrent ( win );
  FindMapping2 ( width, height );
  opti2 = 4;
  redraw2 = true;
} /*ReshapeFunc2*/

void DisplayFunc2 ( GLFWwindow *win )
{
  redraw2 = true;
} /*DisplayFunc2*/

void Redraw2 ( GLFWwindow *win )
{
  glfwMakeContextCurrent ( win );
  if ( opti2 > 0 ) {
    RedrawMyObject2 ();
    redraw2 = false;
  }
  else {
    opti2 --;
    glfwPostEmptyEvent ();
  }
  glfwSwapBuffers ( win );
  ExitIfGLError ( "Redraw2" );
} /*Redraw2*/

void KeyboardFunc2 ( GLFWwindow *win, int key, int scancode, int action, int mods )
{
/*printf ( "key = %d, scancode = %d, action = %d, mods = %d\n", key, scancode, action, mods  );*/
  queue_was_empty = false;
  if ( action == GLFW_PRESS || action == GLFW_REPEAT ) {
    switch ( key ) {
  case GLFW_KEY_ESCAPE:
      glfwSetWindowShouldClose ( window1, true );
      glfwSetWindowShouldClose ( window2, true );
      break;
  default:
      break;
    }
  }
} /*KeyboardFunc2*/

void CharFunc2 ( GLFWwindow *win, unsigned int charcode )
{
/*printf ( "charcode = %d, (%c)\n", charcode, charcode );*/
  queue_was_empty = false;
  switch ( charcode ) {
case '+':
    redraw2 |= PaletteDegElev ();
    break;
case '-':
    redraw2 |= PaletteDegRed ();
    break;
case 'R':  case 'r':
    PaletteReset ();
    redraw1 = redraw2 = true;
    break;
 default:
    break;
  }
} /*CharFunc2*/

void ScrollFunc2 ( GLFWwindow *win, double x, double y )
{
  queue_was_empty = false;
  switch ( app_state ) {
case STATE_SCROLLING_PALETTE:
    redraw1 |= ChangeMaxIter ( y < 0.0 );
    break;
default:
    redraw1 |= PaletteRhoChange ( y < 0.0 );
    break;
  }
} /*ScrollFunc2*/

void MouseFunc2 ( GLFWwindow *win, int button, int action, int mods )
{
  queue_was_empty = false;
  if ( button == GLFW_MOUSE_BUTTON_LEFT ) {
    if ( action == GLFW_PRESS ) {
      glfwGetCursorPos ( win, &last_xi2, &last_eta2 );
      if ( PickBezPoint1 ( last_xi2, last_eta2 ) )
        app_state = STATE_MOVING_CP1;
      else
        app_state = STATE_SCROLLING_PALETTE;
    }
    else
      app_state = STATE_NOTHING;
  }
} /*MouseFunc2*/

void MotionFunc2 ( GLFWwindow *win, double x, double y )
{
  queue_was_empty = false;
  switch ( app_state ) {
case STATE_MOVING_CP1:
    MoveBezPoint1 ( x, y );
    redraw1 = redraw2 = true;
    break;
default:
    break;
  }
  last_xi2 = x;  last_eta2 = y;
} /*MotionFunc2*/

void Cleanup ( void )
{
  DeleteMyObject1 ();
  DeleteMyObject2 ();
#ifdef DEBUG_BUFFERS_ALLOCATION
DumpBufferIdentifiers ();
#endif
} /*Cleanup*/

void Initialise ( int argc, char **argv )
{
  int w, h;

  glfwSetErrorCallback ( myGLFWErrorHandler );
  if ( !glfwInit () ) {
    fprintf ( stderr, "Error: glfwInit failed\n" );
    exit ( 1 );
  }
  if ( !(window1 = glfwCreateWindow ( w = 480, h = 360,
                                      "Mandelbrot set", NULL, NULL )) ) {
    glfwTerminate ();
    ExitOnError ( "Initialise 1" );
  }
  glfwMakeContextCurrent ( window1 );
  GetGLProcAddresses ( 4, 4 );
  PrintGLVersion ();
  InitMyObject1 ();
  InitMyObject2 ();
  glfwSetWindowSizeCallback ( window1, ReshapeFunc1 );
  glfwSetWindowRefreshCallback ( window1, DisplayFunc1 );
  glfwSetKeyCallback ( window1, KeyboardFunc1 );
  glfwSetCharCallback ( window1, CharFunc1 );
  glfwSetMouseButtonCallback ( window1, MouseFunc1 );
  glfwSetCursorPosCallback ( window1, MotionFunc1 );
  glfwSetScrollCallback ( window1, ScrollFunc1 );
  ReshapeFunc1 ( window1, w, h );
  glfwWindowHint ( GLFW_SAMPLES, 8 );
  if ( !(window2 = glfwCreateWindow ( w, h, "Mandelbrot set colours",
                                      NULL, window1 )) ) {
    glfwTerminate ();
    ExitOnError ( "Initialise 1" );
  }
  glfwMakeContextCurrent ( window2 );
  glfwSetWindowSizeCallback ( window2, ReshapeFunc2 );
  glfwSetWindowRefreshCallback ( window2, DisplayFunc2 );
  glfwSetKeyCallback ( window2, KeyboardFunc2 );
  glfwSetCharCallback ( window2, CharFunc2 );
  glfwSetMouseButtonCallback ( window2, MouseFunc2 );
  glfwSetCursorPosCallback ( window2, MotionFunc2 );
  glfwSetScrollCallback ( window2, ScrollFunc2 );
  ReshapeFunc2 ( window2, w, h );
  glfwMakeContextCurrent ( window1 );
} /*Initialise*/

void MessageLoop ( void )
{
int i;

  do {
    queue_was_empty = true;
    glfwWaitEvents ();
i = 1;
    while ( !queue_was_empty ) {
      queue_was_empty = true;
      glfwPollEvents ();
i ++;
    }
/*if ( i > 1 ) printf ( "i = %d\n", i );*/
    if ( redraw1 )
      Redraw1 ( window1 );
    if ( redraw2 )
      Redraw2 ( window2 );
  } while ( !glfwWindowShouldClose ( window1 ) );
} /*MessageLoop*/

int main ( int argc, char **argv )
{
  Initialise ( argc, argv );
  MessageLoop ();
  Cleanup ();
  glfwTerminate ();
  exit ( 0 );
} /*main*/
