#version 420

#define PROJ_LINEAR          0
#define PROJ_POINT_PANORAMA  1
#define PROJ_LINEAR_PANORAMA 2
#define PROJ_SPHERICAL       3

layout(lines) in;
layout(line_strip,max_vertices=81) out;

in Vertex {
  vec4 Colour;
} In[];

out NVertex {
  vec4 Colour;
} Out;

uniform TransBlock {
  mat4 mm, mmti, vm, pm;
  vec4 eyepos;
} trb;

uniform NLProjection {
  int   type;
  float tantheta;
} nlp;

vec4 NonlinTransformation ( vec4 p )
{
  float R;

  switch ( nlp.type ) {
case PROJ_POINT_PANORAMA:
    R = length ( p.xz );
    return vec4 ( atan ( p.x, -p.z ), p.y/R, -R, p.w );
case PROJ_LINEAR_PANORAMA:
    R = length ( p.xz );
    return vec4 ( atan ( p.x, -p.z ), p.y + R*nlp.tantheta, -R, p.w );
case PROJ_SPHERICAL:
    R = length ( p.xyz );
    return vec4 ( p.x/R, p.y/R, -R/p.w, 1.0 );
default:
    return p;
  }
} /*NonlinTransformation*/

void main ( void )
{
  int  i;
  vec4 p;

  for ( i = 0; i <= 80; i++ ) {
    p = mix ( gl_in[0].gl_Position, gl_in[1].gl_Position, float(i)/30.0 );
    gl_Position = trb.pm*(NonlinTransformation ( trb.vm*(trb.mm*p) ));
    Out.Colour = In[0].Colour;
    EmitVertex ();
  }
  EndPrimitive ();
} /*main*/
