#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <sys/times.h> 
#include <math.h>
#include <openglheader.h>

#include "utilities.h"
#include "lights.h"
#include "nonlin.h"

GLuint shader_id[6];
GLuint program_id[2];

GLuint trbbp,       /* dostep do przeksztalcen */
       nltbuf, nltbbp;
GLint  trbsize, trbofs[5],
       nltbsize, nltbofs[2];

GLuint tetra_vao, tetra_vbo[3],
       cube_vao, cube_vbo[3],
       octa_vao, octa_vbo[3],
       icos_vao, icos_vbo[3],
       dodec_vao, dodec_vbo[3],
       table_vao, table_vbo[3];

GLfloat mmatr[6][16];

float   model_rot_axis[3] = {1.0,0.0,0.0};
double  model_rot_angle0 = 1.047, model_rot_angle;
TransBl trans;
LightBl light;
char    proj_type = PROJ_LINEAR;

void LoadMyShaders ( void )
{
  static const char *filename[] =
    { "nonlin0.glsl.vert", "nonlin0.glsl.geom", "nonlin0.glsl.frag",
      "nonlin1.glsl.vert", "nonlin1.glsl.geom", "nonlin1.glsl.frag" };
  static const GLuint shtype[6] =
    { GL_VERTEX_SHADER, GL_GEOMETRY_SHADER, GL_FRAGMENT_SHADER,
      GL_VERTEX_SHADER, GL_GEOMETRY_SHADER, GL_FRAGMENT_SHADER };
  static const GLchar *UTBNames[] =
    { "TransBlock", "TransBlock.mm", "TransBlock.mmti", "TransBlock.vm",
      "TransBlock.pm", "TransBlock.eyepos" };
  static const GLchar *UNLTNames[] =
    { "NLProjection", "NLProjection.type", "NLProjection.tantheta" };
  static const GLchar *ULSNames[] =
    { "LSBlock", "LSBlock.nls", "LSBlock.mask", "LSBlock.hlen",
      "LSBlock.hl.zenith", "LSBlock.hl.sky", "LSBlock.hl.ground",
      "LSBlock.ls[0].ambient", "LSBlock.ls[0].direct", "LSBlock.ls[0].position",
      "LSBlock.ls[0].attenuation", "LSBlock.ls[1].ambient" };
  GLint i;

  for ( i = 0; i < 6; i++ )
    shader_id[i] = CompileShaderFiles ( shtype[i], 1, &filename[i] );
  program_id[0] = LinkShaderProgram ( 3, shader_id, "0" );
  program_id[1] = LinkShaderProgram ( 3, &shader_id[3], "1" );
  GetAccessToUniformBlock ( program_id[0], 5, &UTBNames[0],
                            &trbsize, trbofs, &trbbp );
  GetAccessToUniformBlock ( program_id[0], 2, &UNLTNames[0],
                            &nltbsize, nltbofs, &nltbbp );
  GetAccessToUniformBlock ( program_id[0], 11, &ULSNames[0],
                            &lsbsize, lsbofs, &lsbbp );
  AttachUniformBlockToBP ( program_id[1], UTBNames[0], trbbp );
  AttachUniformBlockToBP ( program_id[1], UNLTNames[0], nltbbp );
  trans.trbuf = NewUniformBuffer ( trbsize, trbbp );
  nltbuf = NewUniformBuffer ( nltbsize, nltbbp );
  light.lsbuf = NewUniformBuffer ( lsbsize, lsbbp );
  ExitIfGLError ( "LoadMyShaders" );
} /*LoadMyShaders*/

void SetupProjection ( int width, int height, GLint type )
{
  float bt, lr, alpha, phi, tanalpha, tantheta;

  glViewport ( 0, 0, width, height );
  glBindBuffer ( GL_UNIFORM_BUFFER, nltbuf );
  glBufferSubData ( GL_UNIFORM_BUFFER, nltbofs[0], sizeof(GLint), &type );
  tantheta = 0.0;
  glBufferSubData ( GL_UNIFORM_BUFFER, nltbofs[1], sizeof(GLfloat), &tantheta );
  alpha = asin ( RR/VDIST );  tanalpha = tan ( alpha );
  switch ( type ) {
case PROJ_LINEAR:
    bt = NEAR * tanalpha;
    lr = bt*(float)width/(float)height;
    M4x4Frustumf ( trans.pm, NULL, -lr, lr, -bt, bt, NEAR, FAR );
    left = -(right = lr);  bottom = -(top = bt);
    break;
case PROJ_POINT_PANORAMA:
    phi = tanalpha*(float)width/(float)height;
    M4x4Orthof ( trans.pm, NULL, -phi, phi, -tanalpha, tanalpha, NEAR, FAR );
    left = -(right = phi);  bottom = -(top = tanalpha);
    break;
case PROJ_LINEAR_PANORAMA:
    phi = tanalpha*(float)width/(float)height;
    bt = DDIST*phi*(float)height/(float)width;
    M4x4Orthof ( trans.pm, NULL, -phi, phi, -bt, bt, NEAR, FAR );
    left = -(right = phi);  bottom = -(top = bt);
    break;
case PROJ_SPHERICAL:
    phi = asin ( RR/VDIST*(float)width/(float)height );
printf ( "alpha = %6.2f, phi =%6.2f\n", alpha*180.0/3.1415926, phi*180.0/3.1415926 );
    M4x4Orthof ( trans.pm, NULL, -sin(phi), sin(phi), -sin(alpha), sin(alpha), NEAR, FAR );
    left = -(right = phi);  bottom = -(top = alpha);
    break;
  }
  near = NEAR;  far = FAR;
  glBindBuffer ( GL_UNIFORM_BUFFER, trans.trbuf );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[3], 16*sizeof(GLfloat), trans.pm );
  win_width = width,  win_height = height;
  ExitIfGLError ( "SetupProjection" );
} /*SetupProjection*/

void SetupModelMatrix ( float axis[3], float angle, GLfloat mat[16] )
{
#define SCF 0.33
  GLfloat ms[16], mr[16], mt[16], ma[16];

  M4x4Scalef ( ms, SCF, SCF, SCF );
  M4x4Translatef ( mt, 0.0, 0.0, 0.0 );
  M4x4Multf ( ma, mt, ms );
  M4x4RotateVf ( mr, axis[0], axis[1], axis[2], angle );
  M4x4Multf ( trans.mm, mr, ma );
  if ( mat ) {
    M4x4Multf ( ma, trans.mm, mat );
    memcpy ( trans.mm, ma, 16*sizeof(GLfloat) );
  }
  glBindBuffer ( GL_UNIFORM_BUFFER, trans.trbuf );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[0], 16*sizeof(GLfloat), trans.mm );
  memcpy ( trans.mmti, trans.mm, 16*sizeof(GLfloat) );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[1], 16*sizeof(GLfloat), trans.mmti );
  ExitIfGLError ( "SetupModelMatrix" );
} /*SetupModelMatrix*/

void InitViewMatrix ( void )
{
  float rm[16], tm[16];

  M4x4Translatef ( tm, -viewer_pos0[0], -viewer_pos0[1], -viewer_pos0[2] );
  M4x4RotateVf ( rm, viewer_rvec[0], viewer_rvec[1], viewer_rvec[2],
                 -viewer_rangle );
  M4x4Multf ( trans.vm, tm, rm );
  glBindBuffer ( GL_UNIFORM_BUFFER, trans.trbuf );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[2], 16*sizeof(GLfloat), trans.vm );
  M4x4Transposef ( tm, rm );
  M4x4MultMVf ( trans.eyepos, tm, viewer_pos0 );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[4], 4*sizeof(GLfloat), trans.eyepos );
  ExitIfGLError ( "InitViewMatrix" );
} /*InitViewMatrix*/

void RotateViewer ( double delta_xi, double delta_eta )
{
  float   vi[3], lgt, vk[3];
  double  angi, angk;
  GLfloat tm[16], rm[16];

  if ( delta_xi == 0 && delta_eta == 0 )
    return;  /* natychmiast uciekamy - nie chcemy dzielic przez 0 */
  vi[0] = (float)delta_eta*(right-left)/(float)win_height;
  vi[1] = (float)delta_xi*(top-bottom)/(float)win_width;
  vi[2] = 0.0;
  lgt = sqrt ( V3DotProductf ( vi, vi ) );
  vi[0] /= lgt;
  vi[1] /= lgt;
  angi = -0.052359878;  /* -3 stopnie */
  V3CompRotationsf ( vk, &angk, viewer_rvec, viewer_rangle, vi, angi );
  memcpy ( viewer_rvec, vk, 3*sizeof(float) );
  viewer_rangle = angk;
  M4x4Translatef ( tm, -viewer_pos0[0], -viewer_pos0[1], -viewer_pos0[2] );
  M4x4RotateVf ( rm, viewer_rvec[0], viewer_rvec[1], viewer_rvec[2],
                 -viewer_rangle );
  M4x4Multf ( trans.vm, tm, rm );
  glBindBuffer ( GL_UNIFORM_BUFFER, trans.trbuf );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[2], 16*sizeof(GLfloat), trans.vm );
  M4x4Transposef ( tm, rm );
  M4x4MultMVf ( trans.eyepos, tm, viewer_pos0 );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[4], 4*sizeof(GLfloat), trans.eyepos );
  ExitIfGLError ( "RotateViewer" );
} /*RotateViewer*/

void ConstructTetrahedronVAO ( void )
{
#define R 0.57735027
  static const GLfloat vertpos[4][3] =
    {{-R,-R,-R},{R,R,-R},{R,-R,R},{-R,R,R}};
  static const GLubyte vertcol[4][3] =
    {{255,0,0},{0,255,0},{0,0,255},{255,255,0}};
  static const GLubyte vertind[15] =
    {0,1,2,3,0,     /* lamana */
     1,3,2,0,       /* 2 odcinki */
     0,1,2,3,0,1};  /* tasma */

  glGenVertexArrays ( 1, &tetra_vao );
  glBindVertexArray ( tetra_vao );
  glGenBuffers ( 3, tetra_vbo );  
  glBindBuffer ( GL_ARRAY_BUFFER, tetra_vbo[0] );
  glBufferData ( GL_ARRAY_BUFFER,
                 4*3*sizeof(GLfloat), vertpos, GL_STATIC_DRAW );
  glEnableVertexAttribArray ( 0 );
  glVertexAttribPointer ( 0, 3, GL_FLOAT, GL_FALSE,
                          3*sizeof(GLfloat), (GLvoid*)0 );
  glBindBuffer ( GL_ARRAY_BUFFER, tetra_vbo[1] );
  glBufferData ( GL_ARRAY_BUFFER,
                 4*3*sizeof(GLubyte), vertcol, GL_STATIC_DRAW );
  glEnableVertexAttribArray ( 1 );
  glVertexAttribPointer ( 1, 3, GL_UNSIGNED_BYTE, GL_TRUE,
                          3*sizeof(GLubyte), (GLvoid*)0 );
  glBindBuffer ( GL_ELEMENT_ARRAY_BUFFER, tetra_vbo[2] ); 
  glBufferData ( GL_ELEMENT_ARRAY_BUFFER,
                 15*sizeof(GLubyte), vertind, GL_STATIC_DRAW );
  M4x4Translatef ( mmatr[0], -2.5, 0.5, 0.5 );
  ExitIfGLError ( "ConstructTetrahedronVAO" );
#undef R
} /*ConstructTetrahedronVAO*/

void DrawTetrahedron ( GLuint program_id, int opt )
{
  glUseProgram ( program_id );
  glBindVertexArray ( tetra_vao );
  switch ( opt ) {
case 0:    /* wierzcholki */   
    glPointSize ( 5.0 );
    glDrawArrays ( GL_POINTS, 0, 4 );
    break;
case 1:    /* krawedzie */
    glBindBuffer ( GL_ELEMENT_ARRAY_BUFFER, tetra_vbo[2] );
    glDrawElements ( GL_LINE_STRIP, 5,
                     GL_UNSIGNED_BYTE, (GLvoid*)0 );
    glDrawElements ( GL_LINES, 4,
                     GL_UNSIGNED_BYTE, (GLvoid*)(5*sizeof(GLubyte)) );
    break;
default:   /* sciany */
    glBindBuffer ( GL_ELEMENT_ARRAY_BUFFER, tetra_vbo[2] );
    glDrawElements ( GL_TRIANGLE_STRIP, 6,
                     GL_UNSIGNED_BYTE, (GLvoid*)(9*sizeof(GLubyte)) );
    break;
  }
} /*DrawTetrahedron*/

void ConstructCubeVAO ( void )
{
#define R 0.57735027
  static const GLfloat vertpos[8][3] =
    {{-R,-R,-R},{-R,R,-R},{-R,R,R},{-R,-R,R},{R,-R,-R},{R,R,-R},{R,R,R},{R,-R,R}};
  static const GLubyte vertcol[8][3] =
    {{255,0,0},{255,127,0},{255,255,0},{127,255,0},{0,255,0},{0,255,127},
     {0,255,255},{0,127,255}};
  static const GLubyte vertind[32] =
    {0,1,2,3,7,5,4,0,   /* lamana */
     0,3,5,6,0,4,2,6,   /* 4 odcinki */
     0,1,3,2,7,6,4,5,2,6,1,5,0,4,3,7}; /* 2 tasmy */

  glGenVertexArrays ( 1, &cube_vao );
  glBindVertexArray ( cube_vao );
  glGenBuffers ( 3, cube_vbo );  
  glBindBuffer ( GL_ARRAY_BUFFER, cube_vbo[0] );
  glBufferData ( GL_ARRAY_BUFFER,
                 8*3*sizeof(GLfloat), vertpos, GL_STATIC_DRAW );
  glEnableVertexAttribArray ( 0 );
  glVertexAttribPointer ( 0, 3, GL_FLOAT, GL_FALSE,
                          3*sizeof(GLfloat), (GLvoid*)0 );
  glBindBuffer ( GL_ARRAY_BUFFER, cube_vbo[1] );
  glBufferData ( GL_ARRAY_BUFFER,
                 8*3*sizeof(GLubyte), vertcol, GL_STATIC_DRAW );
  glEnableVertexAttribArray ( 1 );
  glVertexAttribPointer ( 1, 3, GL_UNSIGNED_BYTE, GL_TRUE,
                          3*sizeof(GLubyte), (GLvoid*)0 );
  glBindBuffer ( GL_ELEMENT_ARRAY_BUFFER, cube_vbo[2] );  
  glBufferData ( GL_ELEMENT_ARRAY_BUFFER,
                 32*sizeof(GLubyte), vertind, GL_STATIC_DRAW );
  M4x4Translatef ( mmatr[1], -2.5, 2.5, 0.5 );
  ExitIfGLError ( "ConstructCubeVAO" );
#undef R
} /*ConstructCubeVAO*/

void DrawCube ( GLuint program_id, int opt )
{
  glUseProgram ( program_id );
  glBindVertexArray ( cube_vao );
  switch ( opt ) {
case 0:    /* wierzcholki */   
    glPointSize ( 5.0 );
    glDrawArrays ( GL_POINTS, 0, 8 );
    break;
case 1:    /* krawedzie */
    glBindBuffer ( GL_ELEMENT_ARRAY_BUFFER, cube_vbo[2] );
    glDrawElements ( GL_LINE_STRIP, 8,
                     GL_UNSIGNED_BYTE, (GLvoid*)0 );
    glDrawElements ( GL_LINES, 8,
                     GL_UNSIGNED_BYTE, (GLvoid*)(8*sizeof(GLubyte)) );
    break;
default:   /* sciany */
    glBindBuffer ( GL_ELEMENT_ARRAY_BUFFER, cube_vbo[2] );
    glDrawElements ( GL_TRIANGLE_STRIP, 8,
                     GL_UNSIGNED_BYTE, (GLvoid*)(16*sizeof(GLubyte)) );
    glDrawElements ( GL_TRIANGLE_STRIP, 8,
                     GL_UNSIGNED_BYTE, (GLvoid*)(24*sizeof(GLubyte)) );
    break;
  }
} /*DrawCube*/

void ConstructOctahedronVAO ( void )
{
#define R 1.0
  static const GLfloat vertpos[6][3] =
    {{0,0,R},{-R,0,0},{0,-R,0},{R,0,0},{0,R,0},{0,0,-R}};
  static const GLubyte vertcol[6][3] =
    {{255,0,0},{255,127,0},{255,255,0},{127,255,0},{0,255,0},{0,255,127}};
  static const GLubyte vertind[25] =
    {0,1,2,3,4,1,5,3,0,2,5,4,0,  /* lamana */
     0,1,2,3,4,1,5,1,4,3,2,1};   /* 2 wachlarze */

  glGenVertexArrays ( 1, &octa_vao );
  glBindVertexArray ( octa_vao );
  glGenBuffers ( 3, octa_vbo );  
  glBindBuffer ( GL_ARRAY_BUFFER, octa_vbo[0] );
  glBufferData ( GL_ARRAY_BUFFER,
                 6*3*sizeof(GLfloat), vertpos, GL_STATIC_DRAW );
  glEnableVertexAttribArray ( 0 );
  glVertexAttribPointer ( 0, 3, GL_FLOAT, GL_FALSE,
                          3*sizeof(GLfloat), (GLvoid*)0 );
  glBindBuffer ( GL_ARRAY_BUFFER, octa_vbo[1] );
  glBufferData ( GL_ARRAY_BUFFER,
                 6*3*sizeof(GLubyte), vertcol, GL_STATIC_DRAW );
  glEnableVertexAttribArray ( 1 );
  glVertexAttribPointer ( 1, 3, GL_UNSIGNED_BYTE, GL_TRUE,
                          3*sizeof(GLubyte), (GLvoid*)0 );
  glBindBuffer ( GL_ELEMENT_ARRAY_BUFFER, octa_vbo[2] );  
  glBufferData ( GL_ELEMENT_ARRAY_BUFFER,
                 25*sizeof(GLubyte), vertind, GL_STATIC_DRAW );
  M4x4Translatef ( mmatr[2], 0.0, 2.5, 0.5 );
  ExitIfGLError ( "ConstructOctahedronVAO" );
#undef R
} /*ConstructOctahedronVAO*/

void DrawOctahedron ( GLuint program_id, int opt )
{
  glUseProgram ( program_id );
  glBindVertexArray ( octa_vao );
  switch ( opt ) {
case 0:    /* wierzcholki */   
    glPointSize ( 5.0 );
    glDrawArrays ( GL_POINTS, 0, 6 );
    break;
case 1:    /* krawedzie */
    glBindBuffer ( GL_ELEMENT_ARRAY_BUFFER, octa_vbo[2] );
    glDrawElements ( GL_LINE_STRIP, 13,
                     GL_UNSIGNED_BYTE, (GLvoid*)0 );
    break;
default:   /* sciany */
    glBindBuffer ( GL_ELEMENT_ARRAY_BUFFER, octa_vbo[2] );
    glDrawElements ( GL_TRIANGLE_FAN, 6,
                     GL_UNSIGNED_BYTE, (GLvoid*)(13*sizeof(GLubyte)) );
    glDrawElements ( GL_TRIANGLE_FAN, 8,
                     GL_UNSIGNED_BYTE, (GLvoid*)(19*sizeof(GLubyte)) );
    break;
  }
} /*DrawOctahedron*/

void ConstructDodecahedronVAO ( void )
{
#define A 0.35682209
#define B 0.57735027
#define C 0.93417236
  static const GLfloat vertpos[20][3] =
    {{ 0.0, A,-C},{ 0.0,-A,-C},{-B,-B,-B},{-C, 0.0,-A},{-B, B,-B},
     { B, B,-B},{ B,-B,-B},{-A,-C, 0.0},{-C, 0.0, A},{-A, C, 0.0},
     { C, 0.0,-A},{ A,-C, 0.0},{-B,-B, B},{-B, B, B},{ A, C, 0.0},
     { C, 0.0, A},{ B,-B, B},{ 0.0,-A, C},{ 0.0, A, C},{ B, B, B}};
  static const GLubyte vertcol[20][3] =
    {{255,0,0},{255,127,0},{255,255,0},{127,255,0},{0,255,0},{0,255,127},
     {0,255,255},{0,127,255},{0,0,255},{127,0,255},{255,0,255},{255,0,127},
     {0,255,0},{0,255,127}, {0,255,255},{0,127,255},{0,0,255},{127,0,255}, 
     {255,0,255},{255,0,127}};
  static const GLubyte vertind[75] =
    {0,4,3,2,1,6,10,15,19,18,17,16,11,7,12,8,13,9,17,5,0, /* lamana */
     0,1,5,10,14,19,1,18,12,17, /* 5 odcinkow */
     0,1,2,3,4,9,14,5,10,6,1,  /* 4 wachlarze */
     7,2,1,6,11,16,17,12,8,3,2,
     13,9,14,19,18,17,12,8,3,4,9,
     15,19,14,5,10,6,11,16,17,18,19};

  glGenVertexArrays ( 1, &dodec_vao );
  glBindVertexArray ( dodec_vao );
  glGenBuffers ( 3, dodec_vbo );  
  glBindBuffer ( GL_ARRAY_BUFFER, dodec_vbo[0] );
  glBufferData ( GL_ARRAY_BUFFER,
                 20*3*sizeof(GLfloat), vertpos, GL_STATIC_DRAW );
  glEnableVertexAttribArray ( 0 );
  glVertexAttribPointer ( 0, 3, GL_FLOAT, GL_FALSE,
                          3*sizeof(GLfloat), (GLvoid*)0 );
  glBindBuffer ( GL_ARRAY_BUFFER, dodec_vbo[1] );
  glBufferData ( GL_ARRAY_BUFFER,
                 20*3*sizeof(GLubyte), vertcol, GL_STATIC_DRAW );
  glEnableVertexAttribArray ( 1 );
  glVertexAttribPointer ( 1, 3, GL_UNSIGNED_BYTE, GL_TRUE,
                          3*sizeof(GLubyte), (GLvoid*)0 );
  glBindBuffer ( GL_ELEMENT_ARRAY_BUFFER, dodec_vbo[2] ); 
  glBufferData ( GL_ELEMENT_ARRAY_BUFFER,
                 75*sizeof(GLubyte), vertind, GL_STATIC_DRAW );
  M4x4Translatef ( mmatr[3], 2.5, 2.5, 0.5 );
  ExitIfGLError ( "ConstructDodecahedronVAO" );
#undef A
#undef B
#undef C
} /*ConstructDodecahedronVAO*/

void DrawDodecahedron ( GLuint program_id, int opt )
{
  int i;

  glUseProgram ( program_id );
  glBindVertexArray ( dodec_vao );
  switch ( opt ) {
case 0:    /* wierzcholki */   
    glPointSize ( 5.0 );
    glDrawArrays ( GL_POINTS, 0, 20 );
    break;
case 1:    /* krawedzie */
    glBindBuffer ( GL_ELEMENT_ARRAY_BUFFER, dodec_vbo[2] );
    glDrawElements ( GL_LINE_STRIP, 21,
                     GL_UNSIGNED_BYTE, (GLvoid*)0 );
    glDrawElements ( GL_LINES, 10,
                     GL_UNSIGNED_BYTE, (GLvoid*)21 );
    break;
default:   /* sciany */
    glBindBuffer ( GL_ELEMENT_ARRAY_BUFFER, dodec_vbo[2] );
    for ( i = 0; i < 4; i++ )
      glDrawElements ( GL_TRIANGLE_FAN, 11,
                   GL_UNSIGNED_BYTE, (GLvoid*)((31+i*11)*sizeof(GLubyte)) );
    break;
  }
} /*DrawDodecahedron*/

void ConstructIcosahedronVAO ( void )
{
#define A 0.52573115
#define B 0.85065085
  static const GLfloat vertpos[12][3] =
    {{ -A,0.0, -B},{  A,0.0, -B},{0.0, -B, -A},{ -B, -A,0.0},
     { -B,  A,0.0},{0.0,  B, -A},{  A,0.0,  B},{ -A,0.0,  B},
     {0.0, -B,  A},{  B, -A,0.0},{  B,  A,0.0},{0.0,  B,  A}};
  static const GLubyte vertcol[12][3] =
    {{255,0,0},{255,127,0},{255,255,0},{127,255,0},{0,255,0},{0,255,127},
     {0,255,255},{0,127,255},{0,0,255},{127,0,255},{255,0,255},{255,0,127}};
  static const GLubyte vertind[62] =
     { 0, 1, 2, 0, 3, 4, 0, 5, 1, 9, 2, 8, 3, /* lamana, od 0 */
       7, 4, 11, 5, 10, 9, 6, 8, 7, 6, 11, 7,
       1, 10, 6,                              /* lamana, od 25 */
       2, 3, 4, 5, 8, 9, 10, 11,              /* 4 odcinki, od 28 */
       0, 1, 2, 3, 4, 5, 1,                   /* wachlarz, od 36 */ 
       6, 7, 8, 9, 10, 11, 7,                 /* wachlarz, od 43 */ 
       1, 9, 2, 8, 3, 7, 4, 11, 5, 10, 1, 9}; /* tasma, od 50 */    

  glGenVertexArrays ( 1, &icos_vao );
  glBindVertexArray ( icos_vao );
  glGenBuffers ( 3, icos_vbo );  
  glBindBuffer ( GL_ARRAY_BUFFER, icos_vbo[0] );
  glBufferData ( GL_ARRAY_BUFFER,
                 12*3*sizeof(GLfloat), vertpos, GL_STATIC_DRAW );
  glEnableVertexAttribArray ( 0 );
  glVertexAttribPointer ( 0, 3, GL_FLOAT, GL_FALSE,
                          3*sizeof(GLfloat), (GLvoid*)0 );
  glBindBuffer ( GL_ARRAY_BUFFER, icos_vbo[1] );
  glBufferData ( GL_ARRAY_BUFFER,
                 12*3*sizeof(GLubyte), vertcol, GL_STATIC_DRAW );
  glEnableVertexAttribArray ( 1 );
  glVertexAttribPointer ( 1, 3, GL_UNSIGNED_BYTE, GL_TRUE,
                          3*sizeof(GLubyte), (GLvoid*)0 );
  glBindBuffer ( GL_ELEMENT_ARRAY_BUFFER, icos_vbo[2] );  
  glBufferData ( GL_ELEMENT_ARRAY_BUFFER,
                 62*sizeof(GLubyte), vertind, GL_STATIC_DRAW );
  M4x4Translatef ( mmatr[4], 2.5, 0.5, 0.5 );
  ExitIfGLError ( "ConstructIcosahedronVAO" );
#undef A
#undef B
} /*ConstructIcosahedronVAO*/

void DrawIcosahedron ( GLuint program_id, int opt )
{
  glUseProgram ( program_id );
  glBindVertexArray ( icos_vao );
  switch ( opt ) {
case 0:    /* wierzcholki */   
    glPointSize ( 5.0 );
    glDrawArrays ( GL_POINTS, 0, 12 );
    break;
case 1:    /* krawedzie */
    glBindBuffer ( GL_ELEMENT_ARRAY_BUFFER, icos_vbo[2] );
    glDrawElements ( GL_LINE_STRIP, 25,
                     GL_UNSIGNED_BYTE, (GLvoid*)0 );
    glDrawElements ( GL_LINE_STRIP, 3,
                     GL_UNSIGNED_BYTE, (GLvoid*)(25*sizeof(GLubyte)) );
    glDrawElements ( GL_LINES, 8, 
                     GL_UNSIGNED_BYTE, (GLvoid*)(28*sizeof(GLubyte)) );
    break;
default:   /* sciany */
    glBindBuffer ( GL_ELEMENT_ARRAY_BUFFER, icos_vbo[2] );
    glDrawElements ( GL_TRIANGLE_FAN, 7,
                     GL_UNSIGNED_BYTE, (GLvoid*)(36*sizeof(GLubyte)) );
    glDrawElements ( GL_TRIANGLE_FAN, 7,
                     GL_UNSIGNED_BYTE, (GLvoid*)(43*sizeof(GLubyte)) );
    glDrawElements ( GL_TRIANGLE_STRIP, 12,
                     GL_UNSIGNED_BYTE, (GLvoid*)(50*sizeof(GLubyte)) );
    break;
  }
} /*DrawIcosahedron*/

void InitTable ( void )
{
  static const GLfloat vertpos[20][3] =
    {{-1.5,-0.5,0},{-3.5,-0.5,0},{-1.5,1.5,0},{-3.5,3.5,0},{0,1.5,0}, {0,3.5,0},{1.5,1.5,0},{3.5,3.5,0},{1.5,-0.5,0},{3.5,-0.5,0},
     {-1.5,-0.5,-1},{-3.5,-0.5,-1},{-1.5,1.5,-1},{-3.5,3.5,-1},{0,1.5,-1}, {0,3.5,-1},{1.5,1.5,-1},{3.5,3.5,-1},{1.5,-0.5,-1},{3.5,-0.5,-1}};
  static const GLfloat colour[3] = {0.0,0.25,0.8};
  static const GLubyte vertind[42] =
     {0,1,2,3,4,5,6,7,8,9,
      10,11,12,13,14,15,16,17,18,19,
      0,10,1,11,3,13,5,15,7,17,9,19,8,18,6,16,4,14,2,12,0,10};

  glGenVertexArrays ( 1, &table_vao );
  glBindVertexArray ( table_vao );
  glGenBuffers ( 2, table_vbo );  
  glBindBuffer ( GL_ARRAY_BUFFER, table_vbo[0] );
  glBufferData ( GL_ARRAY_BUFFER,
                 20*3*sizeof(GLfloat), vertpos, GL_STATIC_DRAW );
  glEnableVertexAttribArray ( 0 );
  glVertexAttribPointer ( 0, 3, GL_FLOAT, GL_FALSE,
                          3*sizeof(GLfloat), (GLvoid*)0 );
  glVertexAttrib3fv ( 1, colour );
  glBindBuffer ( GL_ELEMENT_ARRAY_BUFFER, table_vbo[1] );
  glBufferData ( GL_ELEMENT_ARRAY_BUFFER,
                 42*sizeof(GLubyte), vertind, GL_STATIC_DRAW );
  M4x4Translatef ( mmatr[5], 0.0, 0.0, -0.5 );
  ExitIfGLError ( "ConstructIcosahedronVAO" );
} /*InitTable*/

void DrawTable ( GLuint program_id )
{
  glUseProgram ( program_id );
  glBindVertexArray ( table_vao );
  glBindBuffer ( GL_ELEMENT_ARRAY_BUFFER, table_vbo[1] );
  glDrawElements ( GL_TRIANGLE_STRIP, 10, GL_UNSIGNED_BYTE, (GLvoid*)0 );
  glDrawElements ( GL_TRIANGLE_STRIP, 10, GL_UNSIGNED_BYTE,
                   (GLvoid*)(10*sizeof(GLubyte)) );
  glDrawElements ( GL_TRIANGLE_STRIP, 22, GL_UNSIGNED_BYTE,
                   (GLvoid*)(20*sizeof(GLubyte)) );
} /*DrawTable*/

void DrawGrid ( GLuint program_id )
{
  glUseProgram ( program_id );
  glBindVertexArray ( table_vao );
  glDrawArraysInstanced ( GL_LINES, 0, 2, 62 );
} /*DrawGrid*/

void InitMyObject ( void )
{
  TimerInit ();
  SetupModelMatrix ( model_rot_axis, model_rot_angle, NULL );
  InitViewMatrix ();
  ConstructTetrahedronVAO ();
  ConstructCubeVAO ();
  ConstructOctahedronVAO ();
  ConstructDodecahedronVAO ();
  ConstructIcosahedronVAO ();
  InitTable ();
  InitLights ();
} /*InitMyObject*/

void DeleteMyObject ( void )
{
} /*DeleteMyObject*/

void InitLights ( void )
{
  GLfloat zenith[3] = { 0.0, 0.0, 1.0 };
  GLfloat sky[4]    = { 0.2, 0.2, 0.2, 1.0 };
  GLfloat ground[4] = { 0.03, 0.03, 0.03, 1.0 };
  GLfloat amb0[3] = { 0.1, 0.1, 0.15 };
  GLfloat dir0[3] = { 0.8, 0.8, 0.8 };
  GLfloat pos0[4] = { 0.3, -0.5, 1.0, 0.0 };
  GLfloat atn0[3] = { 1.0, 0.0, 0.0 };

  SetHemisphericalLight ( &light, zenith, sky, ground );
  SetHemisphericalLightOnOff ( &light, 1 );
  SetLightAmbient ( &light, 0, amb0 );
  SetLightDirect ( &light, 0, dir0 );
  SetLightPosition ( &light, 0, pos0 );
  SetLightAttenuation ( &light, 0, atn0 );
  SetLightOnOff ( &light, 0, true );
} /*InitLights*/

void DrawMyScene ( void )
{
  SetupModelMatrix ( model_rot_axis, model_rot_angle, mmatr[0] );
  DrawTetrahedron ( program_id[0], 2 );
  SetupModelMatrix ( model_rot_axis, model_rot_angle, mmatr[1] );
  DrawCube ( program_id[0], 2 );
  SetupModelMatrix ( model_rot_axis, model_rot_angle, mmatr[2] );
  DrawOctahedron ( program_id[0], 2 );
  SetupModelMatrix ( model_rot_axis, model_rot_angle, mmatr[3] );
  DrawDodecahedron ( program_id[0], 2 );
  SetupModelMatrix ( model_rot_axis, model_rot_angle, mmatr[4] );
  DrawIcosahedron ( program_id[0], 2 );
  SetupModelMatrix ( model_rot_axis, model_rot_angle, mmatr[5] );
  DrawTable ( program_id[0] );
  DrawGrid ( program_id[1] );
  DrawGrid ( program_id[1] );
  DrawGrid ( program_id[1] );
  DrawGrid ( program_id[1] );
} /*DrawMyScene*/

