#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "openglheader.h" /* najpierw ten */
#include <GL/freeglut.h>  /* potem ten */

#include "utilities.h"
#include "lights.h"
#include "trans.h"
#include "pyramid.h"

#define MIN_ZOOM 0.2
#define MAX_ZOOM 5.0
#define ZOOM_FCT 1.05

int     WindowHandle;
int     win_width, win_height;
int     last_xi, last_eta;
float   left, right, bottom, top, near, far;
float   zoom = 1.0;
int     zoomexp = 0;

int     app_state = STATE_NOTHING;
float   viewer_rvec[3] = {1.0,0.0,0.0};
double  viewer_rangle = 0.0;

int     ono = 0;
int     level[2] = { 1, 1 };
char    animate = false;

int     opti = 0;

void ReshapeFunc ( int width, int height )
{
  float lr;

  glViewport ( 0, 0, win_width = width, win_height = height );
  lr = zoom*0.5533*(float)width/(float)height;  /* przyjmujemy aspekt rowny 1 */
  M4x4Frustumf ( trans.pm, NULL, left = -lr, right = lr,
                 bottom = -zoom*0.5533, top = zoom*0.5533, near = 5.0, far = 15.0 );
  LoadVPMatrix ( &trans );
  opti = 4;
} /*ReshapeFunc*/

void DisplayFunc ( void )
{
  if ( opti > 0 ) {
    opti --;
    glutPostWindowRedisplay ( WindowHandle );
  }
  else {
    glClearColor ( 1.0, 1.0, 1.0, 1.0 );
    glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    glEnable ( GL_DEPTH_TEST );
    if ( clip )
      DrawClippedSponge ();
    else
      DrawTheObject ( ono );
    glUseProgram ( 0 );
    glFlush ();
  }
  glutSwapBuffers ();
} /*DisplayFunc*/

char ChangeZoom ( char inczoom )   
{
  int   e;
  double z;

  e = zoomexp;
  if ( inczoom ) {
    if ( (z = pow ( ZOOM_FCT, (double)(++e) )) > MAX_ZOOM )
      return false;
  }
  else {
    if ( (z = pow ( ZOOM_FCT, (double)(--e) )) < MIN_ZOOM )
      return false;
  }
  zoomexp = e;  zoom = z;
  ReshapeFunc ( win_width, win_height );
  return true;
} /*ChangeZoom*/

void ToggleAnimation ( void )
{
  if ( (animate = !animate) ) {
    TimerTic ();
    glutIdleFunc ( IdleFunc );
  }
  else
    glutIdleFunc ( NULL );
} /*ToggleAnimation*/

void KeyboardFunc ( unsigned char key, int x, int y )
{
  int maxl, i;

  switch ( key ) {
case 0x1B:            /* klawisz Esc - zatrzymanie programu */
    Cleanup ();
    glutLeaveMainLoop ();
    break;
case '0':
    if ( ono != 0 ) {
      ono = 0;
      glutSetWindowTitle ( "piramida" );
      glutPostWindowRedisplay ( WindowHandle );
    }
    break;
case '1':
    if ( ono != 1 ) {
      ono = 1;
      glutSetWindowTitle ( "gabka" );
      glutPostWindowRedisplay ( WindowHandle );
    }
    break;
case '-':
    if ( level[ono] > MIN_LEVEL ) {
      level[ono] --;
      glutPostWindowRedisplay ( WindowHandle );
    }
    break;
case '+':
    maxl = ono == 0 ? MAX_LEVEL_P : MAX_LEVEL_M;
    if ( level[ono] < maxl ) {
      level[ono] ++;
      glutPostWindowRedisplay ( WindowHandle );
    }
    break;
case ' ':             /* wlaczamy albo wylaczamy animacje */
    ToggleAnimation ();
    break;
case 'L':  case 'l':
    for ( i = 0; i < 3; i++ ) {
      glUseProgram ( program_id[i] );
      glUniform1i ( LightingModelLoc[i], lighting_model = !lighting_model );
    }
    glutPostWindowRedisplay ( WindowHandle );
    break;
case 'C':  case 'c':
    clip = !clip;
    glutPostWindowRedisplay ( WindowHandle );
    break;
default:              /* ignorujemy wszystkie inne klawisze */ 
    break;
  }
} /*KeyboardFunc*/

void MouseFunc ( int button, int state, int x, int y )
{
  switch ( button ) {
case GLUT_LEFT_BUTTON+3:
    if ( state == GLUT_DOWN && ChangeZoom ( true ) )
      glutPostWindowRedisplay ( WindowHandle );
    return;
case GLUT_LEFT_BUTTON+4:
    if ( state == GLUT_DOWN && ChangeZoom ( false ) )
      glutPostWindowRedisplay ( WindowHandle );
    return;
  }
  switch ( app_state ) {
case STATE_NOTHING:
    if ( button == GLUT_LEFT_BUTTON && state == GLUT_DOWN ) {
      last_xi = x,  last_eta = y;
      app_state = STATE_TURNING;
    }
    break;
case STATE_TURNING:
    if ( button == GLUT_LEFT_BUTTON && state != GLUT_DOWN )
      app_state = STATE_NOTHING;
    break;
default:
    break;
  }
} /*MouseFunc*/

void MotionFunc ( int x, int y )
{
  switch ( app_state ) {
case STATE_TURNING:
    if ( x != last_xi || y != last_eta ) {
      RotateViewer ( x-last_xi, y-last_eta );
      last_xi = x,  last_eta = y;
      glutPostWindowRedisplay ( WindowHandle );
    }
    break;
default:
    break;
  }
} /*MotionFunc*/

void JoystickFunc ( unsigned int buttonmask, int x, int y, int z ) { }

void IdleFunc ( void )
{
  if ( (model_rot_angle += ANGULAR_VELOCITY * TimerTocTic ()) >= PI )
    model_rot_angle -= 2.0*PI;
  SetupModelMatrix ( model_rot_axis, model_rot_angle );
  glutPostWindowRedisplay ( WindowHandle );
} /*IdleFunc*/

void TimerFunc ( int value ) {  }

void Initialise ( int argc, char *argv[] )
{
  glutInit ( &argc, argv );
  glutInitContextVersion ( 2, 1 );
  glutInitContextFlags ( GLUT_FORWARD_COMPATIBLE );
  glutInitContextProfile ( GLUT_CORE_PROFILE );
  glutSetOption ( GLUT_ACTION_ON_WINDOW_CLOSE,
                  GLUT_ACTION_GLUTMAINLOOP_RETURNS );
  glutSetOption ( GLUT_MULTISAMPLE, 8 );
  glutInitWindowSize ( 480, 360 );
  glutInitDisplayMode ( GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA |
                        GLUT_MULTISAMPLE );
  WindowHandle = glutCreateWindow ( "piramida" );
  if ( WindowHandle < 1 ) {
    fprintf ( stderr, "Error: Could not create a window\n" );
    exit ( 1 );
  }
  GetGLProcAddresses ( 4, 2 );
  PrintGLVersion ();
  glutReshapeFunc ( ReshapeFunc );
  glutDisplayFunc ( DisplayFunc );
  glutKeyboardFunc ( KeyboardFunc );
  glutMouseFunc ( MouseFunc );
  glutMotionFunc ( MotionFunc );
  /*glutJoystickFunc ( JoystickFunc, 16 );*/
  /*glutTimerFunc ( 0, TimerFunc, 0 );*/
  InitMyObject ();
} /*Initialise*/

int main ( int argc, char *argv[] )
{
  Initialise ( argc, argv );
  glutMainLoop ();
  exit ( 0 );
} /*main*/
