#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include "openglheader.h"   /* najpierw ten */
#include <GL/freeglut.h>    /* potem ten */

#include "utilities.h"
#include "lights.h"
#include "trans.h"
#include "pyramid.h"

GLuint  program_id[4];
GLint   levelloc[4], ColourSourceLoc[4];
GLint   LightingModelLoc[4], lighting_model;
GLint   clpsize, clpofs[2];
GLuint  clpbp, clpbuf;
TransBl trans;
LightBl light;
GLuint  matbuf;
char    clip = false;

float   model_rot_axis[3] = {0.0,1.0,0.0};
double  model_rot_angle = 0.0;
const float viewer_pos0[4] = {0.0,0.0,10.0,1.0};

void LoadMyShaders ( void )
{
  static const char *filename[] =
    { "pyramid.vert.glsl", "pyramid.geom.glsl", "app2c.frag.glsl",
      "sponge.geom.glsl", "sponge1.geom.glsl", "sponge2.geom.glsl" };
  static const GLenum shtype[6] =
    { GL_VERTEX_SHADER, GL_GEOMETRY_SHADER, GL_FRAGMENT_SHADER,
      GL_GEOMETRY_SHADER, GL_GEOMETRY_SHADER, GL_GEOMETRY_SHADER };
  static const GLchar *uvnames[] =
    { "level", "ColourSource", "LightingModel" };
  static const GLchar *uvnames1[] =
    { "ClipPlane", "ClipPlane.nv", "ClipPlane.square" };
  GLuint shader_id[6];
  int    i;

  for ( i = 0; i < 6; i++ )
    shader_id[i] = CompileShaderFiles ( shtype[i], 1, &filename[i] );
  program_id[0] = LinkShaderProgram ( 3, &shader_id[0], "0" );
  glDeleteShader ( shader_id[1] );
  shader_id[1] = shader_id[3];
  program_id[1] = LinkShaderProgram ( 3, &shader_id[0], "1" );
  glDeleteShader ( shader_id[1] );
  shader_id[1] = shader_id[4];
  program_id[2] = LinkShaderProgram ( 3, &shader_id[0], "2" );
  glDeleteShader ( shader_id[1] );
  shader_id[1] = shader_id[5];
  program_id[3] = LinkShaderProgram ( 3, &shader_id[0], "3" );
/*PrintProgramResources ( program_id[3], "3" );*/
  for ( i = 0; i < 4; i++ ) {
    levelloc[i] = glGetUniformLocation ( program_id[i], uvnames[0] );
    ColourSourceLoc[i] = glGetUniformLocation ( program_id[i], uvnames[1] );
    LightingModelLoc[i] = glGetUniformLocation ( program_id[i],  uvnames[2] );
  }
  GetAccessToTransBlockUniform ( program_id[0] );
  GetAccessToLightMatUniformBlocks ( program_id[0] );
  for ( i = 1; i < 4; i++ ) {
    AttachUniformTransBlockToBP ( program_id[i] );
    AttachUniformLightMatBlockToBP ( program_id[i] );
  }
  GetAccessToUniformBlock ( program_id[2], 2, uvnames1, &clpsize, clpofs, &clpbp );
  clpbuf = NewUniformBuffer ( clpsize, clpbp );
  AttachUniformBlockToBP ( program_id[3], uvnames1[0], clpbp );
  for ( i = 0; i < 3; i++ )
    glDeleteShader ( shader_id[i] );
  ExitIfGLError ( "LoadMyShaders" );
} /*LoadMyShaders*/

void SetupModelMatrix ( float axis[3], float angle )
{
  M4x4RotateVf ( trans.mm, axis[0], axis[1], axis[2], angle );
  LoadMMatrix ( &trans, NULL );
} /*SetupModelMatrix*/

void InitViewMatrix ( void )
{
  memcpy ( trans.eyepos, viewer_pos0, 4*sizeof(GLfloat) );
  M4x4Translatef ( trans.vm, -viewer_pos0[0], -viewer_pos0[1], -viewer_pos0[2] );
  LoadVPMatrix ( &trans );
} /*InitViewMatrix*/

void RotateViewer ( int delta_xi, int delta_eta )
{
  float   vi[3], lgt, vk[3];
  double  angi, angk;
  GLfloat tm[16];

  if ( delta_xi == 0 && delta_eta == 0 )
    return;  /* natychmiast uciekamy - nie chcemy dzielic przez 0 */
  vi[0] = (float)delta_eta*(right-left)/(float)win_height;
  vi[1] = (float)delta_xi*(top-bottom)/(float)win_width;
  vi[2] = 0.0;
  lgt = sqrt ( V3DotProductf ( vi, vi ) );
  vi[0] /= lgt;
  vi[1] /= lgt;
  angi = -0.052359878;  /* -3 stopnie */
  V3CompRotationsf ( vk, &angk, viewer_rvec, viewer_rangle, vi, angi );
  memcpy ( viewer_rvec, vk, 3*sizeof(float) );
  viewer_rangle = angk;
  M4x4Translatef ( trans.vm, -viewer_pos0[0], -viewer_pos0[1], -viewer_pos0[2] );
  M4x4MRotateVf ( trans.vm, viewer_rvec[0], viewer_rvec[1], viewer_rvec[2],
                  -viewer_rangle );
  M4x4Transposef ( tm, trans.vm );  tm[3] = tm[7] = tm[11] = 0.0;
  M4x4MultMVf ( trans.eyepos, tm, viewer_pos0 );
  LoadVPMatrix ( &trans );
} /*RotateViewer*/

void DrawTheObject ( int ono )
{
  GLint ninst;
  int   i;

  glEnable ( GL_CULL_FACE );
  glCullFace ( GL_FRONT );
  glFrontFace ( GL_CCW );
  glBindVertexArray ( empty_vao );
  glUseProgram ( program_id[ono] );
  glUniform1i ( levelloc[ono], level[ono] );
  glUniform1i ( ColourSourceLoc[ono], 0 );
  ChooseMaterial ( matbuf );
  if ( ono == 0 )
    ninst = 1 << 2*level[ono];
  else {
    for ( i = 1, ninst = 20; i < level[ono]; i++ )
      ninst *= 20;
  }
  glDrawArraysInstanced ( GL_POINTS, 0, 1, ninst );
  glBindVertexArray ( 0 );
  ExitIfGLError ( "DrawTheObject" );
} /*DrawTheObject*/

void DrawClippedSponge ( void )
{
  GLint ninst;
  int   i;

  glBindVertexArray ( empty_vao );
  ChooseMaterial ( matbuf );
  for ( i = 1, ninst = 20; i < level[1]; i++ )
    ninst *= 20;
  glEnable ( GL_CLIP_DISTANCE0 );
  glEnable ( GL_CULL_FACE );
  glCullFace ( GL_FRONT );
  glFrontFace ( GL_CCW );
  glUseProgram ( program_id[2] );
  glUniform1i ( levelloc[2], level[1] );
  glUniform1i ( ColourSourceLoc[2], 0 );
  glDrawArraysInstanced ( GL_POINTS, 0, 1, ninst );
  glDisable ( GL_CLIP_DISTANCE0 );
  glUseProgram ( program_id[3] );
  glUniform1i ( levelloc[3], level[1] );
  glUniform1i ( ColourSourceLoc[3], 0 );
  glDrawArraysInstanced ( GL_POINTS, 0, 1, ninst );
  glBindVertexArray ( 0 );
  ExitIfGLError ( "DrawClippedSponge" );
} /*DrawClippedSponge*/

void InitLights ( void )
{
  GLfloat amb0[3] = { 0.1, 0.1, 0.15 };
  GLfloat dir0[3] = { 0.8, 0.8, 0.8 };
  GLfloat pos0[4] = { 0.0, 1.0, 1.0, 0.0 };
  GLfloat atn0[3] = { 1.0, 0.0, 0.0 };

  SetLightAmbient ( &light, 0, amb0 );
  SetLightDirect ( &light, 0, dir0 );
  SetLightPosition ( &light, 0, pos0 );
  SetLightAttenuation ( &light, 0, atn0 );
  SetLightOnOff ( &light, 0, 1 );
} /*InitLights*/

void InitMyMaterial ( void )
{
  const GLfloat ambr[4]  = { 0.75, 0.6, 0.2, 1.0 };
  const GLfloat diffr[4] = { 0.75, 0.6, 0.2, 1.0 };
  const GLfloat specr[4] = { 0.7, 0.7, 0.6, 1.0 };
  const GLfloat shn = 60.0, wa = 5.0, we = 5.0;

  matbuf = NewUniformMatBlock ();
  SetupMaterial ( matbuf, ambr, diffr, specr, shn, wa, we );
} /*InitMyMaterial*/

void InitClipPlane ( void )
{
#define A 2.0
  GLfloat nv[4] = { -1.0, -1.0, -1.0, 0.0 };
  GLfloat square[4][4];
  float   v1[4] = {1.0,-1.0,0.0,0.0}, v2[4] = {1.0,1.0,-2.0,0.0}, l;
  int     i;

  l = sqrt ( v1[0]*v1[0] + v1[1]*v1[1] + v1[2]*v1[2] );
  v1[0] /= l;  v1[1] /= l;  v1[2] /= l;
  l = sqrt ( v2[0]*v2[0] + v2[1]*v2[1] + v2[2]*v2[2] );
  v2[0] /= l;  v2[1] /= l;  v2[2] /= l;
  for ( i = 0; i < 4; i++ ) {
    square[0][i] = A*v1[i];
    square[1][i] = A*v2[i];
    square[2][i] = -A*v1[i];
    square[3][i] = -A*v2[i];
  }
  glBindBuffer ( GL_UNIFORM_BUFFER, clpbuf );
  glBufferSubData ( GL_UNIFORM_BUFFER, clpofs[0], 4*sizeof(GLfloat), nv );
  glBufferSubData ( GL_UNIFORM_BUFFER, clpofs[1], 16*sizeof(GLfloat), square );
#undef A
} /*InitClipPlane*/

void InitMyObject ( void )
{
  LoadMyShaders ();
  trans.trbuf = NewUniformTransBlock ();
  light.lsbuf = NewUniformLightBlock ();
  InitMyMaterial ();
  TimerInit ();
  SetupModelMatrix ( model_rot_axis, model_rot_angle );
  InitViewMatrix ();
  ConstructEmptyVAO ();
  InitLights ();
  InitClipPlane ();
} /*InitMyObject*/

void Cleanup ( void )
{
  glUseProgram ( 0 );
  glDeleteProgram ( program_id[0] );
  glDeleteProgram ( program_id[1] );
  glDeleteBuffers ( 1, &trans.trbuf );
  glDeleteBuffers ( 1, &light.lsbuf );
  DeleteEmptyVAO ();
  ExitIfGLError ( "Cleanup" );
  glutDestroyWindow ( WindowHandle );
} /*Cleanup*/

