#version 440

#define A 0.57735027  /* sqrt(1/3) */

layout(points) in;
layout(triangle_strip,max_vertices=6) out;

in int instanceID[];

out FVertex {
    vec3 Colour, Position, Normal, TNormal;
  } Out;

uniform TransBlock {
    mat4 mm, mmti, vm, pm, vpm;
    vec4 eyepos;
  } trb;

uniform ClipPlane {
    vec4 nv;
    vec4 square[4];
  } clp;

uniform int level;

const int t[21][3] =
  {{0,0,0},{1,0,0},{2,0,0},{0,1,0},{2,1,0},{0,2,0,},{1,2,0},{2,2,0},
   {0,0,1},{2,0,1},{0,2,1},{2,2,1},
   {0,0,2},{1,0,2},{2,0,2},{0,1,2},{2,1,2},{0,2,2,},{1,2,2},{2,2,2},
   {0,0,0}};

vec3 cp[20];


int SHClip ( vec3 p, vec3 nv, int n, int k, int l )
{
  vec3  s, t;
  float ds, dt;
  int   i, m;
#define OUTPUT(P) { cp[l+m] = P;  m ++; }

  s = cp[k+n-1];  ds = dot ( nv, p-s );
  for ( i = m = 0;  i < n;  i++ ) {
    t = cp[k+i];  dt = dot ( nv, p-t );
    if ( ds > 0.0 ) {
      if ( dt > 0.0 ) OUTPUT ( t )
      else OUTPUT ( mix ( s, t, ds/(ds-dt) ) )
    }
    else {
      if ( dt > 0.0 ) {
        OUTPUT ( mix ( s, t, ds/(ds-dt) ) )
        OUTPUT ( t )
      }
    }
    s = t;  ds = dt;
  }
  return m;
#undef OUTPUT
} /*SHClip*/

void OutputVertex ( vec3 p )
{
  vec4 q;

  Out.Colour = vec3( p.x/(2*A)+0.5, p.y/(2.0*A)+0.5, p.z/(2.0*A)+0.5 );
  Out.Position = (q = trb.mm*vec4(p,1.0)).xyz;
  Out.Normal = Out.TNormal = mat3(trb.mmti) * clp.nv.xyz;
  gl_Position = trb.vpm * q;
  EmitVertex ();
} /*OutputVertex*/

void OutputPolygon ( int n )
{
  int i, k;

  for ( i = 0, k = n-1;  i < k;  i++, k-- ) {
    OutputVertex ( cp[i] );
    OutputVertex ( cp[k] );
  }
  if ( i == k )
    OutputVertex ( cp[i] );
  EndPrimitive ();
} /*OutputPolygon*/

void main ( void )
{
  float p0[3], p1[3], a, c;
  vec3  p[8];
  int   i, j, k, l;
  bool  pos, neg;

  p0[0] = p0[1] = p0[2] = -A;  p1[0] = p1[1] = p1[2] = A;
  for ( i = 0, j = instanceID[0];  i < level;  i++, j /= 20 ) {
    k = j % 20;
    for ( l = 0; l < 3; l++ )
      switch ( t[k][l] ) {
    case 0:
        p1[l] = (2.0*p0[l] + p1[l])/3.0;
        break;
    case 1:
        a = (2.0*p0[l] + p1[l])/3.0;
        p1[l] = (p0[l] + 2.0*p1[l])/3.0;
        p0[l] = a;
        break;
    case 2:
        p0[l] = (p0[l] + 2.0*p1[l])/3.0;
        break;
      }
  }
  p[0] = vec3(p0[0],p0[1],p0[2]);  p[1] = vec3(p1[0],p0[1],p0[2]);
  p[2] = vec3(p1[0],p1[1],p0[2]);  p[3] = vec3(p0[0],p1[1],p0[2]);
  p[4] = vec3(p0[0],p0[1],p1[2]);  p[5] = vec3(p1[0],p0[1],p1[2]);
  p[6] = vec3(p1[0],p1[1],p1[2]);  p[7] = vec3(p0[0],p1[1],p1[2]);
  for ( i = 0, pos = neg = false;  i < 8;  i++ ) {
    c = dot ( clp.nv.xyz, p[i]-clp.square[0].xyz );
    if ( c > 0.0 ) pos = true;
    else if ( c < 0.0 ) neg = true;
  }
  if ( !(pos && neg) )
    return;
  for ( i = 0; i < 4; i++ )
    cp[i] = clp.square[i].xyz;
  i = SHClip ( p[0], vec3(-1.0,0.0,0.0), 4, 0, 10 );
  i = SHClip ( p[1], vec3(1.0,0.0,0.0),  i, 10, 0 );
  i = SHClip ( p[1], vec3(0.0,-1.0,0.0), i, 0, 10 );
  i = SHClip ( p[2], vec3(0.0,1.0,0.0),  i, 10, 0 );
  i = SHClip ( p[1], vec3(0.0,0.0,-1.0), i, 0, 10 );
  i = SHClip ( p[5], vec3(0.0,0.0,1.0),  i, 10, 0 );
  if ( i > 2 )
    OutputPolygon ( i );
} /*main*/

